@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003-2005 Alexandre Ratchov
+ * Copyright (c) 2003-2007 Alexandre Ratchov <alex@caoua.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -40,43 +40,18 @@
 #include "ev.h"
 #include "filt.h"
 #include "pool.h"
+#include "mux.h"
+
+#define TAG_PASS 1
+#define TAG_PENDING 2
+
+/*
+ * timeout for throtteling: 1 tick at 60 bpm
+ */
+#define FILT_TIMO TEMPO_TO_USEC24(120,24)
 
 unsigned filt_debug = 0;
 
-/* ------------------------------------------------------------------ */
-
-	/*
-	 * states are structures used to hold the notes
-	 * which are "on", the values of used controllers
-	 * the bender etc...
-	 *
-	 * since for each midi event, a state can be created
-	 * in realtime, we use a pool.
-	 *
-	 */
-
-struct pool_s state_pool;
-
-void
-state_pool_init(unsigned size) {
-	pool_init(&state_pool, "state", sizeof(struct state_s), size);
-}
-
-void
-state_pool_done(void) {
-	pool_done(&state_pool);
-}
-
-struct state_s *
-state_new(void) {
-	return (struct state_s *)pool_new(&state_pool);
-}
-
-void
-state_del(struct state_s *s) {
-	pool_del(&state_pool, s);
-}
-
 /* ------------------------------------------------------------------ */
 
 unsigned char filt_curve_id[128] = {
@@ -119,7 +94,7 @@ unsigned char filt_curve_inv[128] = {
 
 
 void
-rule_dbg(struct rule_s *o) {
+rule_dbg(struct rule *o) {
 	switch(o->type) {
 	case RULE_DEVDROP:
 		dbg_puts("devdrop");
@@ -181,7 +156,48 @@ rule_dbg(struct rule_s *o) {
 }
 
 void
-rule_chgich(struct rule_s *o, unsigned olddev, unsigned oldch, 
+rule_chgich(struct rule *o, unsigned olddev, unsigned oldch, 
+    unsigned newdev, unsigned newch) {
+	switch(o->type) {
+	case RULE_CHANDROP:
+	case RULE_CHANMAP:
+	case RULE_KEYDROP:
+	case RULE_KEYMAP:
+	case RULE_CTLDROP:
+	case RULE_CTLMAP:
+		if (o->idev == olddev && o->ich == oldch) {
+			o->idev = newdev;
+			o->ich = newch;
+		}
+		break;
+	default:
+		break;
+	}
+}
+
+
+void
+rule_chgidev(struct rule *o, unsigned olddev, unsigned newdev) {
+	switch(o->type) {
+	case RULE_CHANDROP:
+	case RULE_CHANMAP:
+	case RULE_KEYDROP:
+	case RULE_KEYMAP:
+	case RULE_CTLDROP:
+	case RULE_CTLMAP:
+	case RULE_DEVDROP:
+	case RULE_DEVMAP:
+		if (o->idev == olddev) {
+			o->idev = newdev;
+		}
+		break;
+	default:
+		break;
+	}
+}
+
+void
+rule_swapich(struct rule *o, unsigned olddev, unsigned oldch, 
     unsigned newdev, unsigned newch) {
 	switch(o->type) {
 	case RULE_CHANDROP:
@@ -205,53 +221,7 @@ rule_chgich(struct rule_s *o, unsigned olddev, unsigned oldch,
 
 
 void
-rule_chgidev(struct rule_s *o, unsigned olddev, unsigned newdev) {
-	switch(o->type) {
-	case RULE_CHANDROP:
-	case RULE_CHANMAP:
-	case RULE_KEYDROP:
-	case RULE_KEYMAP:
-	case RULE_CTLDROP:
-	case RULE_CTLMAP:
-	case RULE_DEVDROP:
-	case RULE_DEVMAP:
-		if (o->idev == olddev) {
-			o->idev = newdev;
-		} else if (o->idev == newdev) {
-			o->idev = olddev;
-		}
-		break;
-	default:
-		break;
-	}
-}
-
-void
-rule_swapich(struct rule_s *o, unsigned olddev, unsigned oldch, 
-    unsigned newdev, unsigned newch) {
-	switch(o->type) {
-	case RULE_CHANDROP:
-	case RULE_CHANMAP:
-	case RULE_KEYDROP:
-	case RULE_KEYMAP:
-	case RULE_CTLDROP:
-	case RULE_CTLMAP:
-		if (o->idev == newdev && o->ich == newch) {
-			o->ich = oldch;			
-			o->idev = olddev;			
-		} else if (o->idev == olddev && o->ich == oldch) {
-			o->idev = newdev;
-			o->ich = newch;
-		}
-		break;
-	default:
-		break;
-	}
-}
-
-
-void
-rule_swapidev(struct rule_s *o, unsigned olddev, unsigned newdev) {
+rule_swapidev(struct rule *o, unsigned olddev, unsigned newdev) {
 	switch(o->type) {
 	case RULE_CHANDROP:
 	case RULE_CHANMAP:
@@ -274,16 +244,13 @@ rule_swapidev(struct rule_s *o, unsigned olddev, unsigned newdev) {
 
 
 void
-rule_chgoch(struct rule_s *o, unsigned olddev, unsigned oldch, 
+rule_chgoch(struct rule *o, unsigned olddev, unsigned oldch, 
     unsigned newdev, unsigned newch) {
 	switch(o->type) {
 	case RULE_CHANMAP:
 	case RULE_KEYMAP:
 	case RULE_CTLMAP:
-		if (o->odev == newdev && o->och == newch) {
-			o->och = oldch;			
-			o->odev = olddev;			
-		} else if (o->odev == olddev && o->och == oldch) {
+		if (o->odev == olddev && o->och == oldch) {
 			o->odev = newdev;
 			o->och = newch;
 		}
@@ -295,7 +262,7 @@ rule_chgoch(struct rule_s *o, unsigned olddev, unsigned oldch,
 
 
 void
-rule_chgodev(struct rule_s *o, unsigned olddev, unsigned newdev) {
+rule_chgodev(struct rule *o, unsigned olddev, unsigned newdev) {
 	switch(o->type) {
 	case RULE_CHANMAP:
 	case RULE_KEYMAP:
@@ -303,8 +270,6 @@ rule_chgodev(struct rule_s *o, unsigned olddev, unsigned newdev) {
 	case RULE_DEVMAP:
 		if (o->odev == olddev) {
 			o->odev = newdev;
-		} else if (o->odev == newdev) {
-			o->odev = olddev;
 		}
 		break;
 	default:
@@ -313,7 +278,7 @@ rule_chgodev(struct rule_s *o, unsigned olddev, unsigned newdev) {
 }
 
 void
-rule_swapoch(struct rule_s *o, unsigned olddev, unsigned oldch, 
+rule_swapoch(struct rule *o, unsigned olddev, unsigned oldch, 
     unsigned newdev, unsigned newch) {
 	switch(o->type) {
 	case RULE_CHANMAP:
@@ -334,7 +299,7 @@ rule_swapoch(struct rule_s *o, unsigned olddev, unsigned oldch,
 
 
 void
-rule_swapodev(struct rule_s *o, unsigned olddev, unsigned newdev) {
+rule_swapodev(struct rule *o, unsigned olddev, unsigned newdev) {
 	switch(o->type) {
 	case RULE_CHANMAP:
 	case RULE_KEYMAP:
@@ -353,31 +318,28 @@ rule_swapodev(struct rule_s *o, unsigned olddev, unsigned newdev) {
 
 /* --------------------------------------------------------------------- */
 
+void filt_timocb(void *);
 
-	/* 
-	 * initialise a filter
-	 */
-	 
+/* 
+ * initialise a filter
+ */
 void
-filt_init(struct filt_s *o) {
-	o->cb = NULL;
-	o->addr = NULL;
-		
-	o->statelist = NULL;
+filt_init(struct filt *o) {
+	o->ops = NULL;
+	o->addr = NULL;	
 	o->active = 1;
 	o->voice_rules = NULL;
 	o->chan_rules = NULL;
 	o->dev_rules = NULL;
+	timo_set(&o->timo, filt_timocb, o);
 }
 
-	/*
-	 * removes all filtering rules and all states
-	 */
-
+/*
+ * removes all filtering rules and all states
+ */
 void
-filt_reset(struct filt_s *o) {
-	struct state_s *i;
-	struct rule_s *r;
+filt_reset(struct filt *o) {
+	struct rule *r;
 	while (o->voice_rules) {
 		r = o->voice_rules;
 		o->voice_rules = r->next;
@@ -393,75 +355,64 @@ filt_reset(struct filt_s *o) {
 		o->dev_rules = r->next;
 		mem_free(r); 
 	}
-	while (o->statelist) {
-		i = o->statelist;
-		o->statelist = i->next;
-		state_del(i);
-	}
 }
 
-	/*
-	 * destroy a filter
-	 */
-
+/*
+ * destroy a filter
+ */
 void
-filt_done(struct filt_s *o) {
+filt_done(struct filt *o) {
 #ifdef FILT_DEBUG
-	if (o->cb != NULL) {
+	if (o->ops != NULL) {
 		dbg_puts("filt_done: call filt_stop first.\n");
 	}
-	if (o->statelist != NULL) {
-		dbg_puts("filt_done: statelist not empty\n");
-	}
 #endif
 	filt_reset(o);
 }
 
 
-	/*
-	 * attach a filter to the owener
-	 * events are passed to the given callback (see filt_pass)
-	 */
-
+/*
+ * configure a filter so that output events are passed to the given
+ * callback (see filt_processev)
+ */
 void
-filt_start(struct filt_s *o, void (*cb)(void *, struct ev_s *), void *addr) {
-	/*
-	XXX: why ?
-	o->active = 0; 
-	*/
+filt_start(struct filt *o, struct muxops *ops, void *addr) {
 	o->addr = addr;
-	o->cb = cb;
+	o->ops = ops;
+	statelist_init(&o->statelist);
+	timo_add(&o->timo, FILT_TIMO);
+	if (filt_debug) {
+		dbg_puts("filt_start()\n");
+	}
 }
 
 
-	/*
-	 * detaches the filter from the ower
-	 */
-
+/*
+ * unconfigure a filter
+ */
 void
-filt_stop(struct filt_s *o) {
-	o->cb = NULL;
-	o->addr = NULL;	
-#ifdef FILT_DEBUG
-	if (o->statelist) {
-		dbg_puts("filt_stop: state list isn't empty\n");
+filt_stop(struct filt *o) {
+	if (filt_debug) {
+		dbg_puts("filt_stop()\n");
 	}
-#endif
+	timo_del(&o->timo);
+	statelist_done(&o->statelist);
+	o->ops = NULL;
+	o->addr = NULL;	
 }
 
 
 /* ------------------------------------- configuration management --- */
 
-	/* 
-	 * configure the filt to drop 
-	 * events from a particular device 
-	 * - if the same drop-rule exists then do nothing
-	 * - remove map-rules with the same idev
-	 */
-	 
+/* 
+ * configure the filt to drop 
+ * events from a particular device 
+ * - if the same drop-rule exists then do nothing
+ * - remove map-rules with the same idev
+ */
 void
-filt_conf_devdrop(struct filt_s *o, unsigned idev) {
-	struct rule_s **i, *r;
+filt_conf_devdrop(struct filt *o, unsigned idev) {
+	struct rule **i, *r;
 	
 	i = &o->dev_rules;
 	while (*i) {
@@ -475,21 +426,20 @@ filt_conf_devdrop(struct filt_s *o, unsigned idev) {
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_DEVDROP;
 	r->idev = idev;
 	r->next = o->dev_rules;
 	o->dev_rules = r;
 }
 
-	/* 
-	 * configure the filt not to drop 
-	 * events from a particular device 
-	 */ 
-
+/* 
+ * configure the filt not to drop 
+ * events from a particular device 
+ */ 
 void
-filt_conf_nodevdrop(struct filt_s *o, unsigned idev) {
-	struct rule_s **i, *r;
+filt_conf_nodevdrop(struct filt *o, unsigned idev) {
+	struct rule **i, *r;
 	
 	i = &o->dev_rules;
 	while (*i) {
@@ -504,17 +454,16 @@ filt_conf_nodevdrop(struct filt_s *o, unsigned idev) {
 	}
 }
 
-	/* 
-	 * configure the filter to map one device to another 
-	 * - if the same map-rule exists dont add new one
-	 * - remove maps with the same output dev
-	 * - remove drops with the same input
-	 */
-
+/* 
+ * configure the filter to map one device to another 
+ * - if the same map-rule exists dont add new one
+ * - remove maps with the same output dev
+ * - remove drops with the same input
+ */
 void
-filt_conf_devmap(struct filt_s *o,
+filt_conf_devmap(struct filt *o,
     unsigned idev, unsigned odev) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->dev_rules;
 	while (*i != NULL) {
@@ -529,7 +478,7 @@ filt_conf_devmap(struct filt_s *o,
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_DEVMAP;
 	r->idev = idev;
 	r->odev = odev;
@@ -537,14 +486,13 @@ filt_conf_devmap(struct filt_s *o,
 	o->dev_rules = r;
 }
 
-	/* 
-	 * configure the filter not to map 
-	 * any devices to the given one
-	 */
-
+/* 
+ * configure the filter not to map 
+ * any devices to the given one
+ */
 void
-filt_conf_nodevmap(struct filt_s *o, unsigned odev) {
-	struct rule_s **i, *r;
+filt_conf_nodevmap(struct filt *o, unsigned odev) {
+	struct rule **i, *r;
 	
 	i = &o->dev_rules;
 	while (*i != NULL) {
@@ -559,14 +507,13 @@ filt_conf_nodevmap(struct filt_s *o, unsigned odev) {
 	}
 }
 
-	/*
-	 * configure the filter to drop a particular 
-	 * channel
-	 */
-
+/*
+ * configure the filter to drop a particular 
+ * channel
+ */
 void
-filt_conf_chandrop(struct filt_s *o, unsigned idev, unsigned ich) {
-	struct rule_s **i, *r;
+filt_conf_chandrop(struct filt *o, unsigned idev, unsigned ich) {
+	struct rule **i, *r;
 	
 	i = &o->chan_rules;
 	while (*i) {
@@ -583,7 +530,7 @@ filt_conf_chandrop(struct filt_s *o, unsigned idev, unsigned ich) {
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_CHANDROP;
 	r->idev = idev;
 	r->ich = ich;
@@ -591,14 +538,13 @@ filt_conf_chandrop(struct filt_s *o, unsigned idev, unsigned ich) {
 	o->chan_rules = r;	
 }
 
-	/*
-	 * configure the filter to drop a particular 
-	 * channel
-	 */
-	 
+/*
+ * configure the filter to drop a particular 
+ * channel
+ */
 void
-filt_conf_nochandrop(struct filt_s *o, unsigned idev, unsigned ich) {
-	struct rule_s **i, *r;
+filt_conf_nochandrop(struct filt *o, unsigned idev, unsigned ich) {
+	struct rule **i, *r;
 	
 	i = &o->chan_rules;
 	while (*i) {
@@ -615,9 +561,9 @@ filt_conf_nochandrop(struct filt_s *o, unsigned idev, unsigned ich) {
 }
 
 void
-filt_conf_chanmap(struct filt_s *o, unsigned idev, unsigned ich,
+filt_conf_chanmap(struct filt *o, unsigned idev, unsigned ich,
     unsigned odev, unsigned och) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->chan_rules;
 	while (*i != NULL) {
@@ -634,7 +580,7 @@ filt_conf_chanmap(struct filt_s *o, unsigned idev, unsigned ich,
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_CHANMAP;
 	r->idev = idev;
 	r->ich = ich;
@@ -645,8 +591,8 @@ filt_conf_chanmap(struct filt_s *o, unsigned idev, unsigned ich,
 }
 
 void
-filt_conf_nochanmap(struct filt_s *o, unsigned odev, unsigned och ) {
-	struct rule_s **i, *r;
+filt_conf_nochanmap(struct filt *o, unsigned odev, unsigned och ) {
+	struct rule **i, *r;
 	
 	i = &o->chan_rules;
 	while (*i != NULL) {
@@ -664,9 +610,9 @@ filt_conf_nochanmap(struct filt_s *o, unsigned odev, unsigned och ) {
 
 
 void
-filt_conf_ctldrop(struct filt_s *o, 
+filt_conf_ctldrop(struct filt *o, 
     unsigned idev, unsigned ich, unsigned ictl) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i) {
@@ -686,7 +632,7 @@ filt_conf_ctldrop(struct filt_s *o,
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_CTLDROP;
 	r->idev = idev;
 	r->ich = ich;
@@ -696,9 +642,9 @@ filt_conf_ctldrop(struct filt_s *o,
 }
 
 void
-filt_conf_noctldrop(struct filt_s *o, 
+filt_conf_noctldrop(struct filt *o, 
     unsigned idev, unsigned ich, unsigned ictl) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i) {
@@ -716,10 +662,10 @@ filt_conf_noctldrop(struct filt_s *o,
 }
 
 void
-filt_conf_ctlmap(struct filt_s *o, 
+filt_conf_ctlmap(struct filt *o, 
     unsigned idev, unsigned ich, unsigned odev, unsigned och, 
     unsigned ictl, unsigned octl) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i != NULL) {
@@ -739,7 +685,7 @@ filt_conf_ctlmap(struct filt_s *o,
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_CTLMAP;
 	r->idev = idev;
 	r->ich = ich;
@@ -753,9 +699,9 @@ filt_conf_ctlmap(struct filt_s *o,
 }
 
 void
-filt_conf_noctlmap(struct filt_s *o, 
+filt_conf_noctlmap(struct filt *o, 
     unsigned odev, unsigned och, unsigned octl) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i != NULL) {
@@ -773,9 +719,9 @@ filt_conf_noctlmap(struct filt_s *o,
 }
 
 void
-filt_conf_keydrop(struct filt_s *o, unsigned idev, unsigned ich, 
+filt_conf_keydrop(struct filt *o, unsigned idev, unsigned ich, 
     unsigned keylo, unsigned keyhi) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i) {
@@ -797,7 +743,7 @@ filt_conf_keydrop(struct filt_s *o, unsigned idev, unsigned ich,
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_KEYDROP;
 	r->idev = idev;
 	r->ich = ich;
@@ -808,15 +754,14 @@ filt_conf_keydrop(struct filt_s *o, unsigned idev, unsigned ich,
 }
 
 
-	/*
-	 * if there is intersection with existing rules
-	 * then remove them
-	 */
-
+/*
+ * if there is intersection with existing rules
+ * then remove them
+ */
 void
-filt_conf_nokeydrop(struct filt_s *o, unsigned idev, unsigned ich, 
+filt_conf_nokeydrop(struct filt *o, unsigned idev, unsigned ich, 
     unsigned keylo, unsigned keyhi) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i) {
@@ -836,10 +781,10 @@ filt_conf_nokeydrop(struct filt_s *o, unsigned idev, unsigned ich,
 
 
 void
-filt_conf_keymap(struct filt_s *o, 
+filt_conf_keymap(struct filt *o, 
     unsigned idev, unsigned ich, unsigned odev, unsigned och, 
     unsigned keylo, unsigned keyhi, int keyplus) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i) {
@@ -861,7 +806,7 @@ filt_conf_keymap(struct filt_s *o,
 			i = &(*i)->next;
 		}
 	}
-	r = (struct rule_s *)mem_alloc(sizeof(struct rule_s));
+	r = (struct rule *)mem_alloc(sizeof(struct rule));
 	r->type = RULE_KEYMAP;
 	r->idev = idev;
 	r->ich = ich;
@@ -877,9 +822,9 @@ filt_conf_keymap(struct filt_s *o,
 
 
 void
-filt_conf_nokeymap(struct filt_s *o, unsigned odev, unsigned och, 
+filt_conf_nokeymap(struct filt *o, unsigned odev, unsigned och, 
     unsigned keylo, unsigned keyhi) {
-	struct rule_s **i, *r;
+	struct rule **i, *r;
 	
 	i = &o->voice_rules;
 	while (*i) {
@@ -899,9 +844,9 @@ filt_conf_nokeymap(struct filt_s *o, unsigned odev, unsigned och,
 
 
 void
-filt_conf_chgich(struct filt_s *o, 
+filt_conf_chgich(struct filt *o, 
     unsigned olddev, unsigned oldch, unsigned newdev, unsigned newch) {
-	struct rule_s *i;
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_chgich(i, olddev, oldch, newdev, newch);
@@ -913,8 +858,8 @@ filt_conf_chgich(struct filt_s *o,
 
 
 void
-filt_conf_chgidev(struct filt_s *o, unsigned olddev, unsigned newdev) {
-	struct rule_s *i;
+filt_conf_chgidev(struct filt *o, unsigned olddev, unsigned newdev) {
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_chgidev(i, olddev, newdev);
@@ -929,9 +874,9 @@ filt_conf_chgidev(struct filt_s *o, unsigned olddev, unsigned newdev) {
 
 
 void
-filt_conf_swapich(struct filt_s *o, 
+filt_conf_swapich(struct filt *o, 
     unsigned olddev, unsigned oldch, unsigned newdev, unsigned newch) {
-	struct rule_s *i;
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_swapich(i, olddev, oldch, newdev, newch);
@@ -943,8 +888,8 @@ filt_conf_swapich(struct filt_s *o,
 
 
 void
-filt_conf_swapidev(struct filt_s *o, unsigned olddev, unsigned newdev) {
-	struct rule_s *i;
+filt_conf_swapidev(struct filt *o, unsigned olddev, unsigned newdev) {
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_swapidev(i, olddev, newdev);
@@ -958,9 +903,9 @@ filt_conf_swapidev(struct filt_s *o, unsigned olddev, unsigned newdev) {
 }
 
 void
-filt_conf_chgoch(struct filt_s *o, 
+filt_conf_chgoch(struct filt *o, 
     unsigned olddev, unsigned oldch, unsigned newdev, unsigned newch) {
-	struct rule_s *i;
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_chgoch(i, olddev, oldch, newdev, newch);
@@ -972,8 +917,8 @@ filt_conf_chgoch(struct filt_s *o,
 
 
 void
-filt_conf_chgodev(struct filt_s *o, unsigned olddev, unsigned newdev) {
-	struct rule_s *i;
+filt_conf_chgodev(struct filt *o, unsigned olddev, unsigned newdev) {
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_chgodev(i, olddev, newdev);
@@ -988,9 +933,9 @@ filt_conf_chgodev(struct filt_s *o, unsigned olddev, unsigned newdev) {
 
 
 void
-filt_conf_swapoch(struct filt_s *o, 
+filt_conf_swapoch(struct filt *o, 
     unsigned olddev, unsigned oldch, unsigned newdev, unsigned newch) {
-	struct rule_s *i;
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_swapoch(i, olddev, oldch, newdev, newch);
@@ -1002,8 +947,8 @@ filt_conf_swapoch(struct filt_s *o,
 
 
 void
-filt_conf_swapodev(struct filt_s *o, unsigned olddev, unsigned newdev) {
-	struct rule_s *i;
+filt_conf_swapodev(struct filt *o, unsigned olddev, unsigned newdev) {
+	struct rule *i;
 	
 	for (i = o->voice_rules; i != NULL; i = i->next) {
 		rule_swapodev(i, olddev, newdev);
@@ -1018,86 +963,84 @@ filt_conf_swapodev(struct filt_s *o, unsigned olddev, unsigned newdev) {
 
 /* ----------------------------------------------- real-time part --- */
 
-	/* 
-	 * executes the given rule; if event matches the rule then
-	 * pass it and return 1, else retrun 0
-	 */
-	 	
-		
+/* 
+ * executes the given rule; if the event matches the rule then
+ * pass it and return 1, else retrun 0
+ */
 unsigned
-filt_matchrule(struct filt_s *o, struct rule_s *r, struct ev_s *ev) {
-	struct ev_s te;
+filt_matchrule(struct filt *o, struct rule *r, struct ev *ev) {
+	struct ev te;
 		
 	switch(r->type) {
 	case RULE_DEVDROP:
-		if (ev->data.voice.dev == r->idev) {
+		if (ev->dev == r->idev) {
 			goto match_drop;
 		}
 		break;	
 	case RULE_DEVMAP:
-		if (ev->data.voice.dev == r->idev) {
+		if (ev->dev == r->idev) {
 			te = *ev;
-			te.data.voice.dev = r->odev;
+			te.dev = r->odev;
 			goto match_pass;
 		}
 		break;
 	case RULE_CHANDROP:
-		if (ev->data.voice.dev == r->idev &&
-		    ev->data.voice.ch == r->ich) {
+		if (ev->dev == r->idev &&
+		    ev->ch == r->ich) {
 			goto match_drop;
 		}
 		break;
 	case RULE_CHANMAP:
-		if (ev->data.voice.dev == r->idev && 
-		    ev->data.voice.ch == r->ich) {
+		if (ev->dev == r->idev && 
+		    ev->ch == r->ich) {
 			te = *ev;
-			te.data.voice.dev = r->odev;
-			te.data.voice.ch = r->och;
+			te.dev = r->odev;
+			te.ch = r->och;
 			goto match_pass;
 		}
 		break;
 	case RULE_KEYDROP:
 		if (EV_ISNOTE(ev) && 
-		    ev->data.voice.dev == r->idev && 
-		    ev->data.voice.ch == r->ich && 
-		    ev->data.voice.b0 >= r->keylo &&
-		    ev->data.voice.b0 <= r->keyhi) {
+		    ev->dev == r->idev && 
+		    ev->ch == r->ich && 
+		    ev->note_num >= r->keylo &&
+		    ev->note_num <= r->keyhi) {
 			goto match_drop;
 		}
 		break;
 	case RULE_KEYMAP:
 		if (EV_ISNOTE(ev) && 
-		    ev->data.voice.dev == r->idev && 
-		    ev->data.voice.ch == r->ich && 
-		    ev->data.voice.b0 >= r->keylo &&
-		    ev->data.voice.b0 <= r->keyhi) {
+		    ev->dev == r->idev && 
+		    ev->ch == r->ich && 
+		    ev->note_num >= r->keylo &&
+		    ev->note_num <= r->keyhi) {
 			te = *ev;
-			te.data.voice.dev = r->odev;
-			te.data.voice.ch = r->och;
-			te.data.voice.b0 += r->keyplus;
-			te.data.voice.b0 &= 0x7f;
-			te.data.voice.b1 = r->curve[te.data.voice.b1];
+			te.dev = r->odev;
+			te.ch = r->och;
+			te.note_num += r->keyplus;
+			te.note_num &= 0x7f;
+			te.note_vel = te.note_vel;
 			goto match_pass;
 		}
 		break;
 	case RULE_CTLDROP:
-		if (ev->cmd == EV_CTL &&
-		    ev->data.voice.dev == r->idev &&
-		    ev->data.voice.ch == r->ich &&
-		    ev->data.voice.b0 == r->ictl) {
+		if (ev->cmd == EV_XCTL &&
+		    ev->dev == r->idev &&
+		    ev->ch == r->ich &&
+		    ev->ctl_num == r->ictl) {
 			goto match_drop;
 		}
 		break;
 	case RULE_CTLMAP:
-		if (ev->cmd == EV_CTL &&
-		    ev->data.voice.dev == r->idev &&
-		    ev->data.voice.ch == r->ich &&
-		    ev->data.voice.b0 == r->ictl) {
+		if (ev->cmd == EV_XCTL &&
+		    ev->dev == r->idev &&
+		    ev->ch == r->ich &&
+		    ev->ctl_num == r->ictl) {
 			te = *ev;
-			te.data.voice.dev = r->odev;
-			te.data.voice.ch = r->och;
-			te.data.voice.b0 = r->octl;
-			te.data.voice.b1 = r->curve[te.data.voice.b1];
+			te.dev = r->odev;
+			te.ch = r->och;
+			te.ctl_num = r->octl;
+			te.ctl_val = te.ctl_val;
 			goto match_pass;
 		}
 		break;
@@ -1112,9 +1055,7 @@ match_pass:
 		ev_dbg(&te);
 		dbg_puts("\n");
 	}
-	if (o->cb) {
-		o->cb(o->addr, &te);
-	}
+	o->ops->ev(o->addr, &te);
 	return 1;
 match_drop:
 	if (filt_debug) {
@@ -1126,14 +1067,13 @@ match_drop:
 }
 
 
-	/*
-	 * match event against all rules
-	 */
-
+/*
+ * match event against all rules
+ */
 void
-filt_processev(struct filt_s *o, struct ev_s *ev) {
+filt_processev(struct filt *o, struct ev *ev) {
 	unsigned match;
-	struct rule_s *i;
+	struct rule *i;
 
 	match = 0;
 	for (i = o->voice_rules; i != NULL; i = i->next) {
@@ -1153,200 +1093,76 @@ filt_processev(struct filt_s *o, struct ev_s *ev) {
 					ev_dbg(ev);
 					dbg_puts("\n");
 				}
-				if (o->cb) {
-					o->cb(o->addr, ev);
-				}
+				o->ops->ev(o->addr, ev);
 			}
 		}
 	}
 }
 
-	/*
-	 * send state event and delete
-	 * the state if it is no more used
-	 *
-	 * a pointer to the next runable state in the
-	 * list is returned
-	 */
-
-struct state_s **
-filt_staterun(struct filt_s *o, struct state_s **p) {
-	struct state_s *s = *p;
+/*
+ * shuts all notes and restores the default values
+ * of the modified controllers, the bender etc...
+ */
+void
+filt_shut(struct filt *o) {
+	struct state *s, *snext;
+	struct ev ca;
 	
-	if (s->nevents == 0) {
-		filt_processev(o, &s->ev);
-		if (!s->keep) {
-			*p = s->next;
-#ifdef FILT_DEBUG
-		if (filt_debug) {
-			dbg_puts("filt_staterun: deleting: ");
-			ev_dbg(&s->ev);
-			dbg_puts("\n");
-		}
-#endif
-			state_del(s);
-			return p;
-		}
-	}
-	s->nevents++;
-	return &s->next;
-}
-
-	/*
-	 * create and run a new state and return pointer
-	 * to the next runable state
-	 *
-	 * WARNING: the returned pointer is the next runable state!!!
-	 */
-
-struct state_s **
-filt_statecreate(struct filt_s *o, struct ev_s *ev, unsigned keep) {
-	struct state_s *s;
-	
-#ifdef FILT_DEBUG
-	if (filt_debug) {
-		dbg_puts("filt_statecreate: ");
-		ev_dbg(ev);
-		dbg_puts("\n");
-	}
-#endif
-	s = state_new();
-	s->ev = *ev;
-	s->nevents = 0;
-	s->keep = keep;
-	s->next = o->statelist;
-	o->statelist = s;
-	return filt_staterun(o, &o->statelist);
-}
-
-	/*
-	 * update the event and the keep-flag of the given state
-	 * try to send it
-	 */
-
-struct state_s **
-filt_stateupdate(struct filt_s *o, struct state_s **p, struct ev_s *ev, unsigned keep) {
-	(*p)->ev = *ev;
-	(*p)->keep = keep;
-	return filt_staterun(o, p);
-}
-
-	/*
-	 * lookup in the state list for a state
-	 * corresponding to the give event. Ie a
-	 * state with the same note, the same controller,
-	 * bender etc...
-	 */
-
-struct state_s **
-filt_statelookup(struct filt_s *o, struct ev_s *ev) {
-	struct state_s **i;
-	
-	if (EV_ISNOTE(ev)) {
-		for (i = &o->statelist; *i != NULL; i = &(*i)->next) {
-			if (EV_ISNOTE(&(*i)->ev) && 
-			    (*i)->ev.data.voice.dev == ev->data.voice.dev && 
-			    (*i)->ev.data.voice.ch == ev->data.voice.ch && 
-			    (*i)->ev.data.voice.b0 == ev->data.voice.b0) {
-				return i;
-			}
-		}	
-	} else if (ev->cmd == EV_CTL) {
-		for (i = &o->statelist; *i != NULL; i = &(*i)->next) {
-			if ((*i)->ev.cmd == EV_CTL && 
-			    (*i)->ev.data.voice.dev == ev->data.voice.dev && 
-			    (*i)->ev.data.voice.ch == ev->data.voice.ch && 
-			    (*i)->ev.data.voice.b0 == ev->data.voice.b0) {
-				return i;
+	for (s = o->statelist.first; s != NULL; s = snext) {
+		snext = s->next;
+		if (state_cancel(s, &ca)) {
+			if (filt_debug) {
+				dbg_puts("filt_shut: ");
+				ev_dbg(&s->ev);
+				dbg_puts(": cancelled by: ");
+				ev_dbg(&ca);
+				dbg_puts("\n");
 			}
+			filt_processev(o, &ca);
 		}
-	} else if (ev->cmd == EV_BEND) {
-		for (i = &o->statelist; *i != NULL; i = &(*i)->next) {
-			if ((*i)->ev.cmd == EV_BEND &&
-			    (*i)->ev.data.voice.dev == ev->data.voice.dev && 
-			    (*i)->ev.data.voice.ch == ev->data.voice.ch) {
-				return i;
-			}
-		}	
-	} else if (ev->cmd == EV_CAT) {
-		for (i = &o->statelist; *i != NULL; i = &(*i)->next) {
-			if ((*i)->ev.cmd == EV_CAT && 
-			    (*i)->ev.data.voice.dev == ev->data.voice.dev && 
-			    (*i)->ev.data.voice.ch == ev->data.voice.ch) {
-				return i;
-			}
-		}	
-	}
-	return 0;
-}
-
-
-	/*
-	 * for the given state (returned by filt_statelookup
-	 * or filt_statenew), generates and passes an event
-	 * restoring the default midi state. I.e. generate
-	 * NOTEOFFs for NOTEON-states etc...
-	 * 	
-	 */
-
-struct state_s **
-filt_stateshut(struct filt_s *o, struct state_s **p) {
-	struct ev_s *ev = &(*p)->ev;
-	struct state_s *s = *p;
-
-	if (s->ev.cmd == EV_NON || s->ev.cmd == EV_KAT) {
-		ev->cmd = EV_NOFF;
-		ev->data.voice.b0 = s->ev.data.voice.b0;
-		ev->data.voice.b1 = EV_NOFF_DEFAULTVEL;
-		s->keep = 0;
-		return filt_staterun(o, p);
-	} else if (s->ev.cmd == EV_BEND) {
-		ev->data.voice.b0 = EV_BEND_DEFAULTLO;
-		ev->data.voice.b1 = EV_BEND_DEFAULTHI;
-		s->keep = 0;
-		return filt_staterun(o, p);
-	} else if (s->ev.cmd == EV_CAT) {
-		ev->data.voice.b0 = EV_CAT_DEFAULT;
-		s->keep = 0;
-		return filt_staterun(o, p);
-	} else {
-#ifdef FILT_DEBUG
-		if (filt_debug) {
-			dbg_puts("filt_stateshut: deleting: ");
-			ev_dbg(ev);
-			dbg_puts("\n");
-		}
-		*p = s->next;
+		statelist_rm(&o->statelist, s);
 		state_del(s);
-		return p;
-#endif
 	}
 }
 
-
-	/*
-	 * shuts all notes and restores the default values
-	 * of the modified controllers, the bender etc...
-	 */
-
+/*
+ * kill all tagged frames matching the given event (because a bogus
+ * event was received)
+ */
 void
-filt_shut(struct filt_s *o) {
-	struct state_s **p;
-	p = &o->statelist;
-	while (*p) {
-		p = filt_stateshut(o, p);
+filt_kill(struct filt *o, struct ev *ev) {
+	struct state *st, *stnext;
+	struct ev ca;
+
+	for (st = o->statelist.first; st != NULL; st = stnext) {
+		stnext = st->next;
+		if (!state_match(st, ev) ||
+		    !(st->tag & TAG_PASS) ||
+		    st->phase & EV_PHASE_LAST) {
+			continue;
+		}
+		/*
+		 * cancel/untag the frame and change the phase to
+		 * EV_PHASE_LAST, so the state can be deleted if
+		 * necessary
+		 */
+		if (state_cancel(st, &ca)) {
+			filt_processev(o, &ca);
+		}
+		st->phase = EV_PHASE_LAST;
+		st->tag &= ~TAG_PASS;
+		dbg_puts("filt_kill: ");
+		ev_dbg(&st->ev);
+		dbg_puts(": killed\n");
 	}
 }
 
-
-	/*
-	 * give an event to the filter for processing
-	 */
-
+/*
+ * give an event to the filter for processing
+ */
 void
-filt_evcb(struct filt_s *o, struct ev_s *ev) {
-	struct state_s **p;
-	unsigned keep;
+filt_evcb(struct filt *o, struct ev *ev) {
+	struct state *st;
 	
 	if (filt_debug) {
 		dbg_puts("filt_run: ");
@@ -1354,8 +1170,8 @@ filt_evcb(struct filt_s *o, struct ev_s *ev) {
 		dbg_puts("\n");
 	}
 #ifdef FILT_DEBUG
-	if (o->cb == NULL) {
-		dbg_puts("filt_evcb: cb = NULL, bad initialisation\n");
+	if (o->ops == NULL) {
+		dbg_puts("filt_evcb: ops = NULL, bad initialisation\n");
 		dbg_panic();
 	}
 	if (!EV_ISVOICE(ev)) {
@@ -1364,94 +1180,66 @@ filt_evcb(struct filt_s *o, struct ev_s *ev) {
 	}
 #endif
 
-	/* 
-	 * first, sanitize the input; keep state for
-	 * NOTEs and BENDs
+	/*
+	 * create/update state for this event
 	 */
-
-	if (EV_ISNOTE(ev)) {
-		p = filt_statelookup(o, ev);
-		if (ev->cmd == EV_NON) {
-			if (p) {
-				dbg_puts("filt_evcb: noteon: state exists, ignored\n");
-				return;
+	st = statelist_update(&o->statelist, ev);
+	if (st->phase & EV_PHASE_FIRST) {
+		if (st->flags & STATE_NEW)
+			st->nevents = 0;
+		/*
+		 * XXX: have to do something with the rule set here
+		 */
+		st->tag = o->active ? TAG_PASS : 0;
+		if (st->flags & (STATE_BOGUS | STATE_NESTED)) {
+			if (filt_debug) {
+				dbg_puts("filt_evcb: ");
+				ev_dbg(ev);
+				dbg_puts(": bogus/nested frame\n");
 			}
-			if (!o->active) {
-				return;
-			}
-			filt_statecreate(o, ev, 1);
-		} else if (ev->cmd == EV_NOFF) {
-			if (p) {
-				filt_stateupdate(o, p, ev, 0);
-			} else {
-				if (!o->active) {
-					return;
-				}
-				dbg_puts("filt_evcb: noteoff: state doesn't exist, event ignored\n");
-			}
-		} else {
-			filt_stateupdate(o, p, ev, 1);
-		}
-	} else if (ev->cmd == EV_BEND) {
-		keep = (ev->data.voice.b0 != EV_BEND_DEFAULTLO || 
-			ev->data.voice.b1 != EV_BEND_DEFAULTHI) ? 1 : 0;
-		p = filt_statelookup(o, ev);
-		if (!p || 
-		    ev->data.voice.b0 != (*p)->ev.data.voice.b0 ||
-		    ev->data.voice.b1 != (*p)->ev.data.voice.b1) {
-			if (p) {
-				filt_stateupdate(o, p, ev, keep);
-			} else {
-				if (!o->active) {
-					return;
-				}
-				filt_statecreate(o, ev, keep);
-			}
-		}
-	} else if (ev->cmd == EV_CTL) {
-		keep = 1;
-		p = filt_statelookup(o, ev);
-		if (!p) {
-			if (!o->active) {
-				return;
-			}
-			filt_statecreate(o, ev, keep);
-		} else if (ev->data.voice.b1 != (*p)->ev.data.voice.b1) {
-			filt_stateupdate(o, p, ev, keep);
-		}
-	} else if (ev->cmd == EV_CAT) {
-		/* 
-		XXX: not tested => set keep to zero 
-		keep = ev->data.voice.b0 != 0 ? 1 : 0;
-		*/
-		keep = 0;		
-		p = filt_statelookup(o, ev);
-		if (!p) {
-			if (!o->active) {
-				return;
-			}
-			filt_statecreate(o, ev, keep);
-		} else if (ev->data.voice.b0 != (*p)->ev.data.voice.b0) {
-			filt_stateupdate(o, p, ev, keep);
-		}
-	} else {
-		filt_processev(o, ev);
-	}	
-}
-
-void
-filt_timercb(struct filt_s *o) {
-	struct state_s **p;
-	unsigned nevents;
-
-	p = &o->statelist;
-	while (*p) {
-		nevents = (*p)->nevents;
-		(*p)->nevents = 0;
-		if (nevents > 1) {
-			p = filt_staterun(o, p);
-		} else {
-			p = &(*p)->next;
+			filt_kill(o, ev);
 		}
 	}
+
+	/*
+	 * nothing to do with silent (not selected) frames
+	 */
+	if (!(st->tag & TAG_PASS))
+		return;
+
+	/*
+	 * throttling: if we played more than MAXEV 
+	 * events skip this event only if it doesnt change the
+	 * phase of the frame
+	 */
+	if (st->nevents > FILT_MAXEV &&
+	    (st->phase == EV_PHASE_NEXT || 
+	     st->phase == (EV_PHASE_FIRST | EV_PHASE_LAST))) {
+		st->tag |= TAG_PENDING;
+		return;
+	}
+
+	filt_processev(o, &st->ev);
+	st->nevents++;
+}
+
+/*
+ * timeout: outdate all events, so throttling will enable
+ * output again.
+ */
+void
+filt_timocb(void *addr) {
+	struct filt *o = (struct filt *)addr;
+	struct state *i;
+
+	statelist_outdate(&o->statelist);
+	for (i = o->statelist.first; i != NULL; i = i->next) {
+		i->nevents = 0;
+		if (i->tag & TAG_PENDING) {
+			i->tag &= ~TAG_PENDING;
+			filt_processev(o, &i->ev);
+			i->nevents++;
+		}
+	}
+	timo_add(&o->timo, FILT_TIMO);
 }
