@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003-2005 Alexandre Ratchov
+ * Copyright (c) 2003-2007 Alexandre Ratchov <alex@caoua.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -28,6 +28,11 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+/*
+ * this module implements the tree containing interpreter code. Each
+ * node of the tree represents one instruction.
+ */
+ 
 #include "dbg.h"
 #include "str.h"
 #include "data.h"
@@ -35,10 +40,10 @@
 #include "exec.h"
 #include "cons.h"
 
-struct node_s *
-node_new(struct node_vmt_s *vmt, struct data_s *data) {
-	struct node_s *o;
-	o = (struct node_s *)mem_alloc(sizeof(struct node_s));
+struct node *
+node_new(struct node_vmt *vmt, struct data *data) {
+	struct node *o;
+	o = (struct node *)mem_alloc(sizeof(struct node));
 	o->vmt = vmt;
 	o->data = data;
 	o->list = o->next = NULL;
@@ -46,8 +51,8 @@ node_new(struct node_vmt_s *vmt, struct data_s *data) {
 }
 
 void
-node_delete(struct node_s *o) {
-	struct node_s *i, *inext;
+node_delete(struct node *o) {
+	struct node *i, *inext;
 	if (o == NULL) {
 		return;
 	}
@@ -62,18 +67,13 @@ node_delete(struct node_s *o) {
 }
 
 void
-node_dbg(struct node_s *o, unsigned depth) {
+node_dbg(struct node *o, unsigned depth) {
 #define NODE_MAXDEPTH 30
 	static char str[2 * NODE_MAXDEPTH + 1] = "";
-	struct node_s *i;
+	struct node *i;
 	
 	dbg_puts(str);
-	/* warning 'cond ? val1 : val2' is a 'const char *' in gcc */
-	if (o->next) {
-		dbg_puts("+-");
-	} else {
-	 	dbg_puts("\\-");
-	}
+	dbg_puts(o != NULL && o->next != NULL ? "+-" : "\\-");
 	if (o == NULL) {
 		dbg_puts("<EMPTY>\n");
 		return;
@@ -84,12 +84,7 @@ node_dbg(struct node_s *o, unsigned depth) {
 		data_dbg(o->data);
 		dbg_puts(")");
 	}
-	/* warning 'cond ? val1 : val2' is a 'const char *' in gcc */
-	if (depth >= NODE_MAXDEPTH && o->list) { 
-		dbg_puts("[...]\n");
-	} else {
-		dbg_puts("\n");
-	}
+	dbg_puts(depth >= NODE_MAXDEPTH && o->list ? "[...]\n" : "\n");
 	if (depth < NODE_MAXDEPTH) {
 		str[2 * depth] = o->next ? '|' : ' ';
 		str[2 * depth + 1] = ' ';
@@ -104,13 +99,13 @@ node_dbg(struct node_s *o, unsigned depth) {
 
 
 void
-node_insert(struct node_s **n, struct node_s *e) {
+node_insert(struct node **n, struct node *e) {
 	e->next = *n;
 	*n = e;
 }
 
 void
-node_replace(struct node_s **n, struct node_s *e) {
+node_replace(struct node **n, struct node *e) {
 	if (e->list != NULL) {
 		dbg_puts("node_replace: e->list != NULL\n");
 		dbg_panic();
@@ -122,17 +117,15 @@ node_replace(struct node_s **n, struct node_s *e) {
 }
 
 
-	/*
-	 * run a node.
-	 * the following rule must be respected
-	 * 1) node_exec must be called always with *r == NULL
-	 * 2) in statements (slist,...) *r != NULL if and only if RETURN
-	 * 3) in expressions (add, ...) *r == NULL if and only if ERROR
-	 */
-
-
+/*
+ * run a node.
+ * the following rule must be respected
+ * 1) node_exec must be called always with *r == NULL
+ * 2) in statements (slist,...) *r != NULL if and only if RETURN
+ * 3) in expressions (add, ...) *r == NULL if and only if ERROR
+ */
 unsigned
-node_exec(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec(struct node *o, struct exec *x, struct data **r) {
 	unsigned result;
 	if (x->depth == EXEC_MAXDEPTH) {
 		cons_err("too many nested operations");
@@ -149,13 +142,12 @@ node_exec(struct node_s *o, struct exec_s *x, struct data_s **r) {
 	return result;
 }
 
-	/*
-	 * execute an unary operator ( '-', '!', '~')
-	 */
-
+/*
+ * execute an unary operator ( '-', '!', '~')
+ */
 unsigned
-node_exec_unary(struct node_s *o, struct exec_s *x, struct data_s **r, 
-	unsigned (*func)(struct data_s *)) { 
+node_exec_unary(struct node *o, struct exec *x, struct data **r, 
+	unsigned (*func)(struct data *)) { 
 	if (node_exec(o->list, x, r) == RESULT_ERR) {
 		return RESULT_ERR;
 	}
@@ -165,14 +157,13 @@ node_exec_unary(struct node_s *o, struct exec_s *x, struct data_s **r,
 	return RESULT_OK;
 }
 
-	/*
-	 * execute a binary operator
-	 */
-
+/*
+ * execute a binary operator
+ */
 unsigned
-node_exec_binary(struct node_s *o, struct exec_s *x, struct data_s **r,
-	unsigned (*func)(struct data_s *, struct data_s *)) { 
-	struct data_s *lhs;
+node_exec_binary(struct node *o, struct exec *x, struct data **r,
+	unsigned (*func)(struct data *, struct data *)) { 
+	struct data *lhs;
 	if (node_exec(o->list, x, r) == RESULT_ERR) {
 		return RESULT_ERR;
 	}
@@ -191,20 +182,15 @@ node_exec_binary(struct node_s *o, struct exec_s *x, struct data_s **r,
 	return RESULT_OK;
 }
 
-
-/* ------------------------------------------------------------------ */
-
-
-	/*
-	 * execute a procedure definition: just check arguments
-	 * and move the tree into a proc_s structure
-	 */
-
+/*
+ * execute a procedure definition: just check arguments
+ * and move the tree into a proc structure
+ */
 unsigned
-node_exec_proc(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	struct proc_s *p;
-	struct node_s *a;
-	struct name_s *args;
+node_exec_proc(struct node *o, struct exec *x, struct data **r) {
+	struct proc *p;
+	struct node *a;
+	struct name *args;
 
 	args = NULL;
 	for (a = o->list->list; a != NULL; a = a->next) {
@@ -221,7 +207,7 @@ node_exec_proc(struct node_s *o, struct exec_s *x, struct data_s **r) {
 		node_delete(p->code);
 	} else {
 		p = proc_new(o->data->val.ref);
-		name_insert((struct name_s **)&x->procs, (struct name_s *)p);
+		name_insert((struct name **)&x->procs, (struct name *)p);
 	}
 	p->args = args;
 	p->code = o->list->next;
@@ -230,38 +216,36 @@ node_exec_proc(struct node_s *o, struct exec_s *x, struct data_s **r) {
 }
 
 unsigned
-node_exec_alist(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_alist(struct node *o, struct exec *x, struct data **r) {
 	dbg_puts("node_exec_alist should not be executed\n");
 	return RESULT_ERR;
 }
 
-
-	/*
-	 * execute a list of statements
-	 */
-
+/*
+ * execute a list of statements
+ */
 unsigned
-node_exec_slist(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	struct node_s *i;
+node_exec_slist(struct node *o, struct exec *x, struct data **r) {
+	struct node *i;
 	unsigned result;
 	
 	for (i = o->list; i != NULL; i = i->next) {
 		result = node_exec(i, x, r);
-		if (result != RESULT_OK) {
+		if (result != RESULT_OK) {		
+			/* stop on ERR, BREAK, CONTINUE, RETURN, EXIT */
 			return result;
 		}
 	}
 	return RESULT_OK;
 }
 
-
-	/*
-	 * execute a builtin function
-	 */
-
+/*
+ * execute a builtin function
+ * if the function didn't set 'r', then set it to 'nil'
+ */
 unsigned
-node_exec_builtin(struct node_s *o, struct exec_s *x, struct data_s **r) {	
-	if (!((unsigned (*)(struct exec_s *, struct data_s **))
+node_exec_builtin(struct node *o, struct exec *x, struct data **r) {	
+	if (!((unsigned (*)(struct exec *, struct data **))
 	    o->data->val.user)(x, r)) {
 		return RESULT_ERR;
 	}
@@ -271,24 +255,22 @@ node_exec_builtin(struct node_s *o, struct exec_s *x, struct data_s **r) {
 	return RESULT_OK;		
 }
 
-	/*
-	 * return a constant 
-	 */
-	 
+/*
+ * return a constant 
+ */
 unsigned
-node_exec_cst(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_cst(struct node *o, struct exec *x, struct data **r) {
 	*r = data_newnil();
 	data_assign(*r, o->data);
 	return RESULT_OK;
 }
 
-	/*
-	 * return the value of the variable (in the node)
-	 */
-
+/*
+ * return the value of the variable (in the node)
+ */
 unsigned
-node_exec_var(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	struct var_s *v;
+node_exec_var(struct node *o, struct exec *x, struct data **r) {
+	struct var *v;
 	
 	v = exec_varlookup(x, o->data->val.ref);
 	if (v == NULL) {
@@ -300,31 +282,32 @@ node_exec_var(struct node_s *o, struct exec_s *x, struct data_s **r) {
 	return RESULT_OK;
 }
 
-	/*
-	 * ignore the result of the expression
-	 * (used to ignore return values of calls)
-	 */
-
+/*
+ * ignore the result of the expression
+ * (used to ignore return values of calls)
+ */
 unsigned
-node_exec_ignore(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	if (node_exec(o->list, x, r) == RESULT_ERR) {
-		return RESULT_ERR;
+node_exec_ignore(struct node *o, struct exec *x, struct data **r) {
+	unsigned result;
+	
+	result = node_exec(o->list, x, r);
+	if (result == RESULT_ERR || result == RESULT_EXIT) {
+		return result;
 	}
 	data_delete(*r);
 	*r = NULL;
 	return RESULT_OK;
 }
 
-	/*
-	 * call a procedure
-	 */
-
+/*
+ * call a procedure
+ */
 unsigned
-node_exec_call(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	struct proc_s *p;
-	struct var_s  **oldlocals, *newlocals;
-	struct name_s *argn;
-	struct node_s *argv;
+node_exec_call(struct node *o, struct exec *x, struct data **r) {
+	struct proc *p;
+	struct name **oldlocals, *newlocals;
+	struct name *argn;
+	struct node *argv;
 	char *procname_save;
 	unsigned result;
 	
@@ -345,7 +328,7 @@ node_exec_call(struct node_s *o, struct exec_s *x, struct data_s **r) {
 		if (node_exec(argv, x, r) == RESULT_ERR) {
 			goto finish;
 		}
-		var_insert(&newlocals, var_new(argn->str, *r));
+		var_new(&newlocals, argn->str, *r);
 		argv = argv->next;
 		*r = NULL;
 	}
@@ -357,11 +340,14 @@ node_exec_call(struct node_s *o, struct exec_s *x, struct data_s **r) {
 	x->locals = &newlocals;
 	procname_save = x->procname;
 	x->procname = p->name.str;
-	if (node_exec(p->code, x, r) != RESULT_ERR) {
-		if (*r == NULL) {			/* always return something */
+	result = node_exec(p->code, x, r);
+	if (result != RESULT_ERR) {
+		if (*r == NULL) {	/* we always return something */
 			*r = data_newnil();
 		}
-		result = RESULT_OK;
+		if (result != RESULT_EXIT) {
+			result = RESULT_OK;
+		}
 	}
 	x->locals = oldlocals;
 	x->procname = procname_save;
@@ -371,7 +357,7 @@ finish:
 }
 
 unsigned
-node_exec_if(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_if(struct node *o, struct exec *x, struct data **r) {
 	unsigned cond, result;
 	if (node_exec(o->list, x, r) == RESULT_ERR) {
 		return RESULT_ERR;
@@ -394,10 +380,10 @@ node_exec_if(struct node_s *o, struct exec_s *x, struct data_s **r) {
 }
 
 unsigned
-node_exec_for(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_for(struct node *o, struct exec *x, struct data **r) {
 	unsigned result;
-	struct data_s *list, *i;
-	struct var_s *v;
+	struct data *list, *i;
+	struct var *v;
 	if (node_exec(o->list, x, &list) == RESULT_ERR) {
 		return RESULT_ERR;
 	}
@@ -407,8 +393,7 @@ node_exec_for(struct node_s *o, struct exec_s *x, struct data_s **r) {
 	}
 	v = exec_varlookup(x, o->data->val.ref);
 	if (v == NULL) {
-		v = var_new(o->data->val.ref, data_newnil());
-		var_insert(x->locals, v);
+		v = var_new(x->locals, o->data->val.ref, data_newnil());
 	}
 	result = RESULT_OK;
 	for (i = list->val.list; i != NULL; i = i->next) {
@@ -427,25 +412,29 @@ node_exec_for(struct node_s *o, struct exec_s *x, struct data_s **r) {
 }
 
 unsigned 
-node_exec_return(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_return(struct node *o, struct exec *x, struct data **r) {
 	if (node_exec(o->list, x, r) == RESULT_ERR) {
 		return RESULT_ERR;
 	}
 	return RESULT_RETURN;
 }
 
+unsigned 
+node_exec_exit(struct node *o, struct exec *x, struct data **r) {
+	return RESULT_EXIT;
+}
+
 unsigned
-node_exec_assign(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	struct var_s *v;
-	struct data_s *expr;
+node_exec_assign(struct node *o, struct exec *x, struct data **r) {
+	struct var *v;
+	struct data *expr;
 
 	if (node_exec(o->list, x, &expr) == RESULT_ERR) {
 		return RESULT_ERR;
 	}
 	v = exec_varlookup(x, o->data->val.ref);
 	if (v == NULL) {
-		v = var_new(o->data->val.ref, expr);
-		var_insert(x->locals, v);
+		v = var_new(x->locals, o->data->val.ref, expr);
 	} else {
 		data_delete(v->data);
 		v->data = expr;
@@ -453,23 +442,21 @@ node_exec_assign(struct node_s *o, struct exec_s *x, struct data_s **r) {
 	return RESULT_OK;
 }
 
-	/* 
-	 * do nothing
-	 */
-
+/* 
+ * do nothing
+ */
 unsigned
-node_exec_nop(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_nop(struct node *o, struct exec *x, struct data **r) {
 	return RESULT_OK;
 }
 
-	/* 
-	 * built a list from the expression list
-	 */
-
+/* 
+ * built a list from the expression list
+ */
 unsigned
-node_exec_list(struct node_s *o, struct exec_s *x, struct data_s **r) {
-	struct node_s *arg;
-	struct data_s *d;
+node_exec_list(struct node *o, struct exec *x, struct data **r) {
+	struct node *arg;
+	struct data *d;
 	
 	*r = data_newlist(NULL);
 
@@ -485,112 +472,112 @@ node_exec_list(struct node_s *o, struct exec_s *x, struct data_s **r) {
 }
 
 unsigned
-node_exec_eq(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_eq(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_eq);
 }
 
 unsigned
-node_exec_neq(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_neq(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_neq);
 }
 
 unsigned
-node_exec_le(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_le(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_le);
 }
 
 unsigned
-node_exec_lt(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_lt(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_lt);
 }
 
 
 unsigned
-node_exec_ge(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_ge(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_ge);
 }
 
 unsigned
-node_exec_gt(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_gt(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_gt);
 }
 
 unsigned
-node_exec_and(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_and(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_and);
 }
 
 unsigned
-node_exec_or(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_or(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_or);
 }
 
 unsigned
-node_exec_not(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_not(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_unary(o, x, r, data_not);
 }
 
 unsigned
-node_exec_add(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_add(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_add);
 }
 
 unsigned
-node_exec_sub(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_sub(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_sub);
 }
 
 unsigned
-node_exec_mul(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_mul(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_mul);
 }
 
 unsigned
-node_exec_div(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_div(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_div);
 }
 
 unsigned
-node_exec_mod(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_mod(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_mod);
 }
 
 unsigned
-node_exec_neg(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_neg(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_unary(o, x, r, data_neg);
 }
 
 unsigned
-node_exec_lshift(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_lshift(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_lshift);
 }
 
 unsigned
-node_exec_rshift(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_rshift(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_rshift);
 }
 
 unsigned
-node_exec_bitand(struct node_s *o, struct exec_s *x, struct data_s **r) { 
+node_exec_bitand(struct node *o, struct exec *x, struct data **r) { 
 	return node_exec_binary(o, x, r, data_bitand);
 }
 
 unsigned
-node_exec_bitor(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_bitor(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_bitor);
 }
 
 unsigned
-node_exec_bitxor(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_bitxor(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_binary(o, x, r, data_bitxor);
 }
 
 unsigned
-node_exec_bitnot(struct node_s *o, struct exec_s *x, struct data_s **r) {
+node_exec_bitnot(struct node *o, struct exec *x, struct data **r) {
 	return node_exec_unary(o, x, r, data_bitnot);
 }
 
-struct node_vmt_s 
+struct node_vmt 
 node_vmt_proc = { "proc", node_exec_proc },
 node_vmt_alist = { "alist", node_exec_alist },
 node_vmt_slist = { "slist", node_exec_slist },
@@ -602,6 +589,7 @@ node_vmt_builtin = { "builtin", node_exec_builtin },
 node_vmt_if = { "if", node_exec_if },
 node_vmt_for = { "for", node_exec_for },
 node_vmt_return = { "return", node_exec_return },
+node_vmt_exit = { "exit", node_exec_exit },
 node_vmt_assign = { "assign", node_exec_assign },
 node_vmt_nop = { "nop", node_exec_nop },
 node_vmt_list = { "list", node_exec_list},
