@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003-2005 Alexandre Ratchov
+ * Copyright (c) 2003-2007 Alexandre Ratchov <alex@caoua.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -111,7 +111,7 @@
 /* ------------------------------------------------------------- */
 
 void
-parse_recover(struct parse_s *o, char *msg) {
+parse_recover(struct parse *o, char *msg) {
 	lex_err(&o->lex, msg);
 	for (;;) {
 		if (o->lex.id == TOK_EOF) {
@@ -127,10 +127,10 @@ parse_recover(struct parse_s *o, char *msg) {
 	}
 }
 
-struct parse_s *
+struct parse *
 parse_new(char *filename) {
-	struct parse_s *o;
-	o = (struct parse_s *)mem_alloc(sizeof(struct parse_s));	
+	struct parse *o;
+	o = (struct parse *)mem_alloc(sizeof(struct parse));	
 	if (!lex_init(&o->lex, filename)) {
 		mem_free(o);
 		return 0;
@@ -140,13 +140,13 @@ parse_new(char *filename) {
 }
 
 void
-parse_delete(struct parse_s *o) {
+parse_delete(struct parse *o) {
 	lex_done(&o->lex);
 	mem_free(o);
 }
 
 unsigned
-parse_getsym(struct parse_s *o) {
+parse_getsym(struct parse *o) {
 	if (o->lookavail) {
 		o->lookavail = 0;
 		return 1;
@@ -155,7 +155,7 @@ parse_getsym(struct parse_s *o) {
 }
 
 void
-parse_ungetsym(struct parse_s *o) {
+parse_ungetsym(struct parse *o) {
 	if (o->lookavail) {
 		dbg_puts("parse_ungetsym: looksym already set\n");
 		dbg_panic();
@@ -164,7 +164,7 @@ parse_ungetsym(struct parse_s *o) {
 }
 
 unsigned
-parse_isfirst(struct parse_s *o, unsigned *first) {
+parse_isfirst(struct parse *o, unsigned *first) {
 	while (*first != 0) {
 		if (*first == o->lex.id) {
 			return 1;
@@ -200,11 +200,11 @@ unsigned first_expr[] = {
 
 unsigned first_stmt[] = {
 	TOK_SEMICOLON, TOK_ENDLINE, 
-	TOK_IF, TOK_FOR, TOK_IDENT, TOK_LET, TOK_RETURN, 0
+	TOK_IF, TOK_FOR, TOK_IDENT, TOK_LET, TOK_RETURN, TOK_EXIT, 0
 };
 	 
 unsigned
-parse_endl(struct parse_s *o, struct node_s **n) {
+parse_endl(struct parse *o, struct node **n) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -216,8 +216,8 @@ parse_endl(struct parse_s *o, struct node_s **n) {
 }
 
 unsigned
-parse_cst(struct parse_s *o, struct node_s **n) {
-	struct data_s *data;
+parse_cst(struct parse *o, struct node **n) {
+	struct data *data;
 		
 	if (!parse_getsym(o)) {
 		return 0;
@@ -315,7 +315,7 @@ parse_cst(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_unary(struct parse_s *o, struct node_s **n) {
+parse_unary(struct parse *o, struct node **n) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -345,7 +345,7 @@ parse_unary(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_muldiv(struct parse_s *o, struct node_s **n) {
+parse_muldiv(struct parse *o, struct node **n) {
 	if (!parse_unary(o, n)) {
 		return 0;
 	}	
@@ -378,7 +378,7 @@ parse_muldiv(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_addsub(struct parse_s *o, struct node_s **n) {
+parse_addsub(struct parse *o, struct node **n) {
 	if (!parse_muldiv(o, n)) {
 		return 0;
 	}
@@ -406,7 +406,7 @@ parse_addsub(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_shift(struct parse_s *o, struct node_s **n) {
+parse_shift(struct parse *o, struct node **n) {
 	if (!parse_addsub(o, n)) {
 		return 0;
 	}
@@ -434,7 +434,7 @@ parse_shift(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_compare(struct parse_s *o, struct node_s **n) {
+parse_compare(struct parse *o, struct node **n) {
 	if (!parse_shift(o, n)) {
 		return 0;
 	}
@@ -472,7 +472,7 @@ parse_compare(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_equal(struct parse_s *o, struct node_s **n) {
+parse_equal(struct parse *o, struct node **n) {
 	if (!parse_compare(o, n)) {
 		return 0;
 	}
@@ -500,7 +500,7 @@ parse_equal(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_bitand(struct parse_s *o, struct node_s **n) {
+parse_bitand(struct parse *o, struct node **n) {
 	if (!parse_equal(o, n)) {
 		return 0;
 	}
@@ -523,7 +523,7 @@ parse_bitand(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_bitxor(struct parse_s *o, struct node_s **n) {
+parse_bitxor(struct parse *o, struct node **n) {
 	if (!parse_bitand(o, n)) {
 		return 0;
 	}
@@ -546,7 +546,7 @@ parse_bitxor(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_bitor(struct parse_s *o, struct node_s **n) {
+parse_bitor(struct parse *o, struct node **n) {
 	if (!parse_bitxor(o, n)) {
 		return 0;
 	}
@@ -569,7 +569,7 @@ parse_bitor(struct parse_s *o, struct node_s **n) {
 
 
 unsigned
-parse_and(struct parse_s *o, struct node_s **n) {
+parse_and(struct parse *o, struct node **n) {
 	if (!parse_bitor(o, n)) {
 		return 0;
 	}
@@ -591,7 +591,7 @@ parse_and(struct parse_s *o, struct node_s **n) {
 }
 
 unsigned
-parse_or(struct parse_s *o, struct node_s **n) {
+parse_or(struct parse *o, struct node **n) {
 	if (!parse_and(o, n)) {
 		return 0;
 	}
@@ -613,12 +613,12 @@ parse_or(struct parse_s *o, struct node_s **n) {
 }
 
 unsigned
-parse_expr(struct parse_s *o, struct node_s **n) {
+parse_expr(struct parse *o, struct node **n) {
 	return parse_or(o, n);
 }
 
 unsigned
-parse_call(struct parse_s *o, struct node_s **n) {
+parse_call(struct parse *o, struct node **n) {
 	if (!parse_getsym(o)) {
 		return 0;
 	} 
@@ -641,11 +641,11 @@ parse_call(struct parse_s *o, struct node_s **n) {
 		}
 		n = &(*n)->next;
 	}
-	return 0; /* aviod compiler warning */
+	/* not reached */
 }
 
 unsigned
-parse_stmt(struct parse_s *o, struct node_s **n) {
+parse_stmt(struct parse *o, struct node **n) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -755,6 +755,12 @@ parse_stmt(struct parse_s *o, struct node_s **n) {
 			return 0;
 		}
 		return 1;
+	} else if (o->lex.id == TOK_EXIT) {
+		*n = node_new(&node_vmt_exit, NULL);
+		if (!parse_endl(o, NULL)) {
+			return 0;
+		}
+		return 1;
 	} else  if (parse_isfirst(o, first_endl)) {
 		parse_ungetsym(o);
 		*n = node_new(&node_vmt_nop, NULL);
@@ -768,7 +774,7 @@ parse_stmt(struct parse_s *o, struct node_s **n) {
 }
 
 unsigned
-parse_slist(struct parse_s *o, struct node_s **n) {
+parse_slist(struct parse *o, struct node **n) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -795,7 +801,7 @@ parse_slist(struct parse_s *o, struct node_s **n) {
 }
 
 unsigned
-parse_alist(struct parse_s *o, struct node_s **n) {
+parse_alist(struct parse *o, struct node **n) {
 	*n = node_new(&node_vmt_alist, NULL);
 	n = &(*n)->list;
 	for(;;) {
@@ -813,8 +819,8 @@ parse_alist(struct parse_s *o, struct node_s **n) {
 }
 
 unsigned
-parse_proc(struct parse_s *o, struct node_s **n) {	
-	struct node_s **a;
+parse_proc(struct parse *o, struct node **n) {	
+	struct node **a;
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -858,7 +864,7 @@ parse_proc(struct parse_s *o, struct node_s **n) {
 }	
 
 unsigned
-parse_line(struct parse_s *o, struct node_s **n) {
+parse_line(struct parse *o, struct node **n) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -880,7 +886,7 @@ parse_line(struct parse_s *o, struct node_s **n) {
 }
 	
 unsigned
-parse_prog(struct parse_s *o, struct node_s **n) {
+parse_prog(struct parse *o, struct node **n) {
 	*n = node_new(&node_vmt_slist, NULL);
 	n = &(*n)->list;
 	for (;;) {
@@ -896,6 +902,6 @@ parse_prog(struct parse_s *o, struct node_s **n) {
 		}
 		n = &(*n)->next;
 	}
-	return 1;
+	/* not reached */
 }
 
