@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003-2005 Alexandre Ratchov
+ * Copyright (c) 2003-2007 Alexandre Ratchov <alex@caoua.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -35,12 +35,13 @@
 #include "parse.h"
 #include "textio.h"
 #include "saveload.h"
-#include "trackop.h"
+#include "frame.h"
+#include "conv.h"
 
 /* ------------------------------------------------------------------- */
 
 void
-chan_output(unsigned dev, unsigned ch, struct textout_s *f) {
+chan_output(unsigned dev, unsigned ch, struct textout *f) {
 	textout_putstr(f, "{");
 	textout_putlong(f, dev);
 	textout_putstr(f, " ");
@@ -50,11 +51,10 @@ chan_output(unsigned dev, unsigned ch, struct textout_s *f) {
 
 
 void
-ev_output(struct ev_s *e, struct textout_s *f) {
+ev_output(struct ev *e, struct textout *f) {
 
 	/* XXX: use ev_getstr() */
 
-	textout_indent(f);
 	switch(e->cmd) {
 		case EV_NOFF:
 			textout_putstr(f, "noff");
@@ -65,33 +65,55 @@ ev_output(struct ev_s *e, struct textout_s *f) {
 		case EV_CAT:
 			textout_putstr(f, "cat");
 			goto one;
-		case EV_CTL:
-			textout_putstr(f, "ctl");
+		case EV_XCTL:
+			if (e->ctl_num < 32) {
+				textout_putstr(f, "xctl");
+				goto two;
+			} else {
+				textout_putstr(f, "ctl");
+				textout_putstr(f, " ");
+				chan_output(e->dev, e->ch, f);
+				textout_putstr(f, " ");
+				textout_putlong(f, e->ctl_num);
+				textout_putstr(f, " ");
+				textout_putlong(f, e->ctl_val >> 7);
+			}
+			break;
+		case EV_XPC:
+			textout_putstr(f, "xpc");
+			goto two;
+		case EV_RPN:
+			textout_putstr(f, "rpn");
+			goto two;
+		case EV_NRPN:
+			textout_putstr(f, "nrpn");
 			goto two;
-		case EV_PC:
-			textout_putstr(f, "pc");
-			goto one;
 		case EV_KAT:
 			textout_putstr(f, "kat");
 			goto two;
 		case EV_BEND:
 			textout_putstr(f, "bend");
-			goto two;
+			textout_putstr(f, " ");
+			chan_output(e->dev, e->ch, f);
+			textout_putstr(f, " ");
+			textout_putlong(f, e->bend_val & 0x7f);
+			textout_putstr(f, " ");
+			textout_putlong(f, e->bend_val >> 7);
+			break;
 		case EV_TEMPO:
 			textout_putstr(f, "tempo");
 			textout_putstr(f, " ");
-			textout_putlong(f, e->data.tempo.usec24);
-			textout_putstr(f, "\n");
+			textout_putlong(f, e->tempo_usec24);
 			break;			
 		case EV_TIMESIG:
 			textout_putstr(f, "timesig");
 			textout_putstr(f, " ");
-			textout_putlong(f, e->data.sign.beats);
+			textout_putlong(f, e->timesig_beats);
 			textout_putstr(f, " ");
-			textout_putlong(f, e->data.sign.tics);
-			textout_putstr(f, "\n");
+			textout_putlong(f, e->timesig_tics);
 			break;			
 		default:
+			textout_putstr(f, "# ignored event\n");
 			dbg_puts("ignoring event: ");
 			ev_dbg(e);
 			dbg_puts("\n");
@@ -100,45 +122,44 @@ ev_output(struct ev_s *e, struct textout_s *f) {
 	return;	
 two:
 	textout_putstr(f, " ");
-	chan_output(e->data.voice.dev, e->data.voice.ch, f);
+	chan_output(e->dev, e->ch, f);
 	textout_putstr(f, " ");
-	textout_putlong(f, e->data.voice.b0);
+	textout_putlong(f, e->v0);
 	textout_putstr(f, " ");
-	textout_putlong(f, e->data.voice.b1);
-	textout_putstr(f, "\n");
+	if (e->v1 != EV_UNDEF) {
+		textout_putlong(f, e->v1);
+	} else {
+		textout_putstr(f, "nil");
+	}
 	return;
 one:	
 	textout_putstr(f, " ");
-	chan_output(e->data.voice.dev, e->data.voice.ch, f);
+	chan_output(e->dev, e->ch, f);
 	textout_putstr(f, " ");
-	textout_putlong(f, e->data.voice.b0);
-	textout_putstr(f, "\n");
+	textout_putlong(f, e->v0);
 	return;
 }
 	
 
 void
-track_output(struct track_s *t, struct textout_s *f) {
-	unsigned delta;
-	struct ev_s ev;
-	struct seqptr_s tp;
+track_output(struct track *t, struct textout *f) {
+	struct seqev *i;
 	
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
 	
-	track_rew(t, &tp);
-	for (;;) {
-		delta = track_ticlast(t, &tp);
-		if (delta != 0) {
+	for (i = t->first; i != NULL; i = i->next) {
+		if (i->delta != 0) {
 			textout_indent(f);
-			textout_putlong(f, delta);
+			textout_putlong(f, i->delta);
 			textout_putstr(f, "\n");		
 		}
-		if (!track_evavail(t, &tp)) {
+		if (i->ev.cmd == EV_NULL) {
 			break;
 		}
-		track_evget(t, &tp, &ev);
-		ev_output(&ev, f);
+		textout_indent(f);
+		ev_output(&i->ev, f);
+		textout_putstr(f, "\n");		
 	}
 	
 	textout_shiftleft(f);
@@ -148,7 +169,7 @@ track_output(struct track_s *t, struct textout_s *f) {
 
 
 void
-rule_output(struct rule_s *o, struct textout_s *f) {
+rule_output(struct rule *o, struct textout *f) {
 	textout_indent(f);
 	switch(o->type) {
 	case RULE_DEVDROP:
@@ -218,8 +239,8 @@ rule_output(struct rule_s *o, struct textout_s *f) {
 }
 
 void
-filt_output(struct filt_s *o, struct textout_s *f) {
-	struct rule_s *i;
+filt_output(struct filt *o, struct textout *f) {
+	struct rule *i;
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
 	
@@ -239,8 +260,8 @@ filt_output(struct filt_s *o, struct textout_s *f) {
 }
 
 void
-sysex_output(struct sysex_s *o, struct textout_s *f) {
-	struct chunk_s *c;
+sysex_output(struct sysex *o, struct textout *f) {
+	struct chunk *c;
 	unsigned i, col;
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
@@ -278,10 +299,9 @@ sysex_output(struct sysex_s *o, struct textout_s *f) {
 	textout_putstr(f, "}");
 }
 
-
 void
-songsx_output(struct songsx_s *o, struct textout_s *f) {
-	struct sysex_s *i;
+songsx_output(struct songsx *o, struct textout *f) {
+	struct sysex *i;
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
 	
@@ -299,7 +319,7 @@ songsx_output(struct songsx_s *o, struct textout_s *f) {
 
 
 void
-songtrk_output(struct songtrk_s *o, struct textout_s *f) {
+songtrk_output(struct songtrk *o, struct textout *f) {
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
 
@@ -325,7 +345,7 @@ songtrk_output(struct songtrk_s *o, struct textout_s *f) {
 }
 
 void
-songchan_output(struct songchan_s *o, struct textout_s *f) {
+songchan_output(struct songchan *o, struct textout *f) {
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
 			
@@ -352,7 +372,7 @@ songchan_output(struct songchan_s *o, struct textout_s *f) {
 }
 
 void
-songfilt_output(struct songfilt_s *o, struct textout_s *f) {
+songfilt_output(struct songfilt *o, struct textout *f) {
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
 
@@ -374,11 +394,70 @@ songfilt_output(struct songfilt_s *o, struct textout_s *f) {
 }
 
 void
-song_output(struct song_s *o, struct textout_s *f) {
-	struct songtrk_s *t;
-	struct songchan_s *i;
-	struct songfilt_s *g;
-	struct songsx_s *s;
+metro_output(struct metro *o, struct textout *f) {
+	textout_putstr(f, "{\n");
+	textout_shiftright(f);
+	
+	textout_indent(f);
+	textout_putstr(f, "enabled ");
+	textout_putlong(f, o->enabled);
+	textout_putstr(f, "\n");
+	
+	textout_indent(f);
+	textout_putstr(f, "lo\t");
+	ev_output(&o->lo, f);
+	textout_putstr(f, "\n");
+
+	textout_indent(f);
+	textout_putstr(f, "hi\t");
+	ev_output(&o->hi, f);
+	textout_putstr(f, "\n");
+
+	textout_shiftleft(f);
+	textout_indent(f);
+	textout_putstr(f, "}");
+}
+
+void
+evctltab_output(struct evctl *tab, struct textout *f) {
+	unsigned i;
+	struct evctl *ctl;
+	
+	textout_putstr(f, "ctltab {\n");
+	textout_shiftright(f);
+	textout_indent(f);
+	textout_putstr(f, "#\n");
+	textout_indent(f);
+	textout_putstr(f, "# name\tnumber\tdefval\n");
+	textout_indent(f);
+	textout_putstr(f, "#\n");
+	for (i = 0; i < 128; i++) {
+		ctl = &tab[i];
+		if (ctl->name) {
+			textout_indent(f);
+			textout_putstr(f, ctl->name);
+			textout_putstr(f, "\t");
+			textout_putlong(f, i);
+			textout_putstr(f, "\t");
+			if (ctl->defval == EV_UNDEF) {
+				textout_putstr(f, "nil");
+			} else {
+				textout_putlong(f, ctl->defval);
+
+			}
+    			textout_putstr(f, "\n");
+		}
+	}
+	textout_shiftleft(f);
+	textout_putstr(f, "}\n");
+}
+
+void
+song_output(struct song *o, struct textout *f) {
+	struct songtrk *t;
+	struct songchan *i;
+	struct songfilt *g;
+	struct songsx *s;
 
 	textout_putstr(f, "{\n");
 	textout_shiftright(f);
@@ -388,12 +467,17 @@ song_output(struct song_s *o, struct textout_s *f) {
 	textout_putlong(f, o->tics_per_unit);
 	textout_putstr(f, "\n");
 
+	textout_indent(f);
+	textout_putstr(f, "tempo_factor ");
+	textout_putlong(f, o->tempo_factor);
+	textout_putstr(f, "\n");
+
 	textout_indent(f);
 	textout_putstr(f, "meta ");
 	track_output(&o->meta, f);
 	textout_putstr(f, "\n");
 
-	for (i = o->chanlist; i != NULL; i = (struct songchan_s *)i->name.next) {
+	SONG_FOREACH_CHAN(o, i) {
 		textout_indent(f);
 		textout_putstr(f, "songchan ");
 		textout_putstr(f, i->name.str);
@@ -401,7 +485,7 @@ song_output(struct song_s *o, struct textout_s *f) {
 		songchan_output(i, f);
 		textout_putstr(f, "\n");
 	}
-	for (g = o->filtlist; g != NULL; g = (struct songfilt_s *)g->name.next) {
+	SONG_FOREACH_FILT(o, g) {
 		textout_indent(f);
 		textout_putstr(f, "songfilt ");
 		textout_putstr(f, g->name.str);
@@ -409,7 +493,7 @@ song_output(struct song_s *o, struct textout_s *f) {
 		songfilt_output(g, f);
 		textout_putstr(f, "\n");
 	}
-	for (t = o->trklist; t != NULL; t = (struct songtrk_s *)t->name.next) {
+	SONG_FOREACH_TRK(o, t) {
 		textout_indent(f);
 		textout_putstr(f, "songtrk ");
 		textout_putstr(f, t->name.str);
@@ -417,7 +501,7 @@ song_output(struct song_s *o, struct textout_s *f) {
 		songtrk_output(t, f);
 		textout_putstr(f, "\n");
 	}
-	for (s = o->sxlist; s != NULL; s = (struct songsx_s *)s->name.next) {
+	SONG_FOREACH_SX(o, s) {
 		textout_indent(f);
 		textout_putstr(f, "songsx ");
 		textout_putstr(f, s->name.str);
@@ -454,14 +538,13 @@ song_output(struct song_s *o, struct textout_s *f) {
 	textout_putstr(f, "curpos ");
 	textout_putlong(f, o->curpos);
 	textout_putstr(f, "\n");
-	textout_indent(f);
 	
 	textout_indent(f);
 	textout_putstr(f, "curlen ");
 	textout_putlong(f, o->curlen);
 	textout_putstr(f, "\n");
+	
 	textout_indent(f);
-
 	textout_putstr(f, "curquant ");
 	textout_putlong(f, o->curquant);
 	textout_putstr(f, "\n");
@@ -473,6 +556,11 @@ song_output(struct song_s *o, struct textout_s *f) {
 	textout_putlong(f, o->curinput_ch);
 	textout_putstr(f, "}\n");
 
+	textout_indent(f);
+	textout_putstr(f, "metro ");
+	metro_output(&o->metro, f);
+	textout_putstr(f, "\n");
+
 	textout_shiftleft(f);
 	textout_indent(f);
 	textout_putstr(f, "}");
@@ -480,16 +568,16 @@ song_output(struct song_s *o, struct textout_s *f) {
 
 /* ---------------------------------------------------------------------- */
 
-unsigned parse_ukline(struct parse_s *o);
-unsigned parse_ukblock(struct parse_s *o);
-unsigned parse_delta(struct parse_s *o, unsigned *delta);
-unsigned parse_ev(struct parse_s *o, struct ev_s *ev);
-unsigned parse_track(struct parse_s *o, struct track_s *t);
-unsigned parse_songtrk(struct parse_s *o, struct song_s *s, struct songtrk_s *t);
-unsigned parse_song(struct parse_s *o, struct song_s *s);
+unsigned parse_ukline(struct parse *o);
+unsigned parse_ukblock(struct parse *o);
+unsigned parse_delta(struct parse *o, unsigned *delta);
+unsigned parse_ev(struct parse *o, struct ev *ev);
+unsigned parse_track(struct parse *o, struct track *t);
+unsigned parse_songtrk(struct parse *o, struct song *s, struct songtrk *t);
+unsigned parse_song(struct parse *o, struct song *s);
 
 unsigned
-parse_ukline(struct parse_s *o) {
+parse_ukline(struct parse *o) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -518,7 +606,7 @@ parse_ukline(struct parse_s *o) {
 }
 
 unsigned
-parse_ukblock(struct parse_s *o) {
+parse_ukblock(struct parse *o) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -543,7 +631,7 @@ parse_ukblock(struct parse_s *o) {
 }
 
 unsigned
-parse_nl(struct parse_s *o) {
+parse_nl(struct parse *o) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -557,7 +645,7 @@ parse_nl(struct parse_s *o) {
 }
 
 unsigned
-parse_empty(struct parse_s *o) {
+parse_empty(struct parse *o) {
 	for (;;) {
 		if (!parse_getsym(o)) {
 			return 0;
@@ -572,7 +660,7 @@ parse_empty(struct parse_s *o) {
 
 
 unsigned
-parse_long(struct parse_s *o, unsigned long max, unsigned long *data) {
+parse_long(struct parse *o, unsigned long min, unsigned long max, unsigned long *data) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -580,8 +668,8 @@ parse_long(struct parse_s *o, unsigned long max, unsigned long *data) {
 		lex_err(&o->lex, "number expected");
 		return 0;
 	}
-	if (o->lex.longval > max) {
-		lex_err(&o->lex, "number to large");
+	if (o->lex.longval < min || o->lex.longval > max) {
+		lex_err(&o->lex, "number out of allowed range");
 		return 0;
 	}
 	*data = (unsigned)o->lex.longval;
@@ -589,7 +677,7 @@ parse_long(struct parse_s *o, unsigned long max, unsigned long *data) {
 }
 
 unsigned
-parse_delta(struct parse_s *o, unsigned *delta) {
+parse_delta(struct parse *o, unsigned *delta) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -605,15 +693,15 @@ parse_delta(struct parse_s *o, unsigned *delta) {
 }
 
 unsigned
-parse_chan(struct parse_s *o, unsigned long *dev, unsigned long *ch) {
+parse_chan(struct parse *o, unsigned long *dev, unsigned long *ch) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
 	if (o->lex.id == TOK_LBRACE) {
-		if (!parse_long(o, EV_MAXDEV, dev)) {
+		if (!parse_long(o, 0, EV_MAXDEV, dev)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXCH, ch)) {
+		if (!parse_long(o, 0, EV_MAXCH, ch)) {
 			return 0;
 		}
 		if (!parse_getsym(o)) {
@@ -638,9 +726,8 @@ parse_chan(struct parse_s *o, unsigned long *dev, unsigned long *ch) {
 	}
 }
 
-
 unsigned
-parse_ev(struct parse_s *o, struct ev_s *ev) {
+parse_ev(struct parse *o, struct ev *ev) {
 	unsigned long val, val2;
 	
 	if (!parse_getsym(o)) {
@@ -651,82 +738,152 @@ parse_ev(struct parse_s *o, struct ev_s *ev) {
 		return 0;
 	}
 	if (!ev_str2cmd(ev, o->lex.strval)) {
-		goto ignore;
-	}
-	if (EV_ISVOICE(ev)) {
-		if (!parse_chan(o, &val, &val2)) {
-			return 0;
-		}
-		ev->data.voice.dev = val;
-		ev->data.voice.ch = val2;
-		if (!parse_long(o, EV_MAXB0, &val)) {
-			return 0;
-		}
-		ev->data.voice.b0 = val;
-		if (ev->cmd != EV_PC && ev->cmd != EV_CAT) {
-			if (ev->cmd == EV_KAT) {
-				/*
-				 * XXX: midish 0.2.5 used to generate
-				 * bogus kat events (without the last
-				 * byte.  As workaround, we ignore
-				 * such event, in order to allow user
-				 * to load its files
-				 */
-				if (!parse_getsym(o)) {
-					return 0;
-				}
-				if (o->lex.id != TOK_NUM) {
-					parse_ungetsym(o);
-					if (!parse_nl(o)) {
-						return 0;
-					}
-					ev->cmd = EV_NULL;
-					return 1;
-				}
-				parse_ungetsym(o);
-			}
-			if (!parse_long(o, EV_MAXB1, &val)) {
-				return 0;
-			}
-			ev->data.voice.b1 = val;
-		}
-	} else if (ev->cmd == EV_TEMPO) {
-		if (!parse_long(o, ~1U, &val)) {
-			return 0;
-		}
-		ev->data.tempo.usec24 = val;
-	} else if (ev->cmd == EV_TIMESIG) {
-		if (!parse_long(o, ~1U, &val)) {
-			return 0;
-		}
-		ev->data.sign.beats = o->lex.longval;
-		if (!parse_long(o, ~1U, &val)) {
-			return 0;
-		}
-		ev->data.sign.tics = o->lex.longval;
-	} else {
-ignore:		parse_ungetsym(o);
+	ignore:
+		parse_ungetsym(o);
 		if (!parse_ukline(o)) {
 			return 0;
 		}
 		lex_err(&o->lex, "unknown event, ignored");
 		ev->cmd = EV_NULL;
 		return 1;
-		
+	}
+	if (EV_ISVOICE(ev)) {
+		if (!parse_chan(o, &val, &val2)) {
+			return 0;
+		}
+		ev->dev = val;
+		ev->ch = val2;
+	}
+	switch (ev->cmd) {
+	case EV_TEMPO:
+		if (!parse_long(o, TEMPO_MIN, TEMPO_MAX, &val)) {
+			return 0;
+		}
+		ev->tempo_usec24 = val;
+		break;
+	case EV_TIMESIG:
+		if (!parse_long(o, 1, TIMESIG_BEATS_MAX, &val)) {
+			return 0;
+		}
+		ev->timesig_beats = val;
+		if (!parse_long(o, 1, TIMESIG_TICS_MAX, &val)) {
+			return 0;
+		}
+		ev->timesig_tics = val;
+		break;
+	case EV_NRPN:
+	case EV_RPN:
+		if (!parse_long(o, 0, EV_MAXFINE, &val)) {
+			return 0;
+		}
+		ev->rpn_num = val;
+		if (!parse_long(o, 0, EV_MAXFINE, &val)) {
+			return 0;
+		}
+		ev->rpn_val = val;
+		break;
+	case EV_XCTL:
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->ctl_num = val;
+		if (!parse_long(o, 0, EV_MAXFINE, &val)) {
+			return 0;
+		}
+		ev->ctl_val = val;
+		break;
+	case EV_XPC:
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->pc_prog = val;
+		if (!parse_getsym(o)) {
+			return 0;
+		}
+		if (o->lex.id == TOK_NIL) {
+			ev->pc_bank = EV_UNDEF;
+			break;
+		}
+		parse_ungetsym(o);
+		if (!parse_long(o, 0, EV_MAXFINE, &val)) {
+			return 0;
+		}
+		ev->pc_bank = val;
+		break;
+	case EV_NON:
+	case EV_NOFF:
+	case EV_CTL:
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v0 = val;
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v1 = val;
+		break;
+	case EV_KAT:
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v0 = val;
+		/*
+		 * XXX: midish version < 0.2.6 used to
+		 * generate bogus kat events (without the last
+		 * byte). As workaround, we ignore such
+		 * events, in order to allow user to load its
+		 * files. Remove this code when no more needed
+		 */
+		if (!parse_getsym(o)) {
+			return 0;
+		}
+		if (o->lex.id != TOK_NUM) {
+			parse_ungetsym(o);
+			if (!parse_nl(o)) {
+				return 0;
+			}
+			ev->cmd = EV_NULL;
+			return 1;
+		}
+		parse_ungetsym(o);
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v1 = val;
+		break;		
+	case EV_PC:
+	case EV_CAT:
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v0 = val;
+		break;
+	case EV_BEND:
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v0 = val;
+		if (!parse_long(o, 0, EV_MAXCOARSE, &val)) {
+			return 0;
+		}
+		ev->v0 += (val << 7);
+		break;
+	default:
+		goto ignore;
 	}
 	if (!parse_nl(o)) {
 		return 0;
 	}
 	return 1;
 }
-	
 
 
 unsigned
-parse_track(struct parse_s *o, struct track_s *t) {
+parse_track(struct parse *o, struct track *t) {
 	unsigned delta;
-	struct ev_s ev;
-	struct seqptr_s tp;
+	struct seqev *pos, *se;
+	struct statelist slist;
+	struct ev ev, rev;
 	
 	if (!parse_getsym(o)) {
 		return 0;
@@ -735,9 +892,12 @@ parse_track(struct parse_s *o, struct track_s *t) {
 		lex_err(&o->lex, "'{' expected while parsing track");
 		return 0;
 	}
-	track_rew(t, &tp);	
+	track_clear(t);	
+	statelist_init(&slist);
+	pos = t->first;
 	for (;;) {
 		if (!parse_getsym(o)) {
+			statelist_done(&slist);
 			return 0;
 		}
 		if (o->lex.id == TOK_ENDLINE) {
@@ -747,27 +907,32 @@ parse_track(struct parse_s *o, struct track_s *t) {
 		} else if (o->lex.id == TOK_NUM) {
 			parse_ungetsym(o);
 			if (!parse_delta(o, &delta)) {
+				statelist_done(&slist);
 				return 0;
 			}
-			delta = o->lex.longval;
-			track_seekblank(t, &tp, delta);
+			pos->delta += delta;
 		} else {
 			parse_ungetsym(o);
 			if (!parse_ev(o, &ev)) {
+				statelist_done(&slist);
 				return 0;
 			}
 			if (ev.cmd != EV_NULL) {
-				track_evlast(t, &tp);
-				track_evput(t, &tp, &ev);
+				if (conv_packev(&slist, 0U, &ev, &rev)) {
+					se = seqev_new();
+					se->ev = rev;
+					seqev_ins(pos, se);
+				}
 			}
 		}
 	}
+	statelist_done(&slist);
 	return 1;			
 }
 
 
 unsigned
-parse_rule(struct parse_s *o, struct filt_s *f) {
+parse_rule(struct parse *o, struct filt *f) {
 	unsigned long idev, ich, odev, och, ictl, octl, keylo, keyhi, ukeyplus;
 	int keyplus;
 	if (!parse_getsym(o)) {
@@ -781,10 +946,10 @@ parse_rule(struct parse_s *o, struct filt_s *f) {
 		if (!parse_chan(o, &idev, &ich)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &keylo)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &keylo)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &keyhi)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &keyhi)) {
 			return 0;
 		}
 		filt_conf_keydrop(f, idev, ich, keylo, keyhi);
@@ -795,13 +960,13 @@ parse_rule(struct parse_s *o, struct filt_s *f) {
 		if (!parse_chan(o, &odev, &och)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &keylo)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &keylo)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &keyhi)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &keyhi)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &ukeyplus)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &ukeyplus)) {
 			return 0;
 		}
 		if (ukeyplus > 63) {
@@ -821,7 +986,7 @@ parse_rule(struct parse_s *o, struct filt_s *f) {
 		if (!parse_chan(o, &idev, &ich)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &ictl)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &ictl)) {
 			return 0;
 		}
 		filt_conf_ctldrop(f, idev, ich, ictl);
@@ -832,10 +997,10 @@ parse_rule(struct parse_s *o, struct filt_s *f) {
 		if (!parse_chan(o, &odev, &och)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &ictl)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &ictl)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXB0, &octl)) {
+		if (!parse_long(o, 0, EV_MAXCOARSE, &octl)) {
 			return 0;
 		}
 		if (!parse_getsym(o)) {
@@ -860,15 +1025,15 @@ parse_rule(struct parse_s *o, struct filt_s *f) {
 		}
 		filt_conf_chanmap(f, idev, ich, odev, och);
 	} else if (str_eq(o->lex.strval, "devdrop")) {
-		if (!parse_long(o, EV_MAXDEV, &idev)) {
+		if (!parse_long(o, 0, EV_MAXDEV, &idev)) {
 			return 0;
 		}
 		filt_conf_devdrop(f, idev);
 	} else if (str_eq(o->lex.strval, "devmap")) {
-		if (!parse_long(o, EV_MAXDEV, &idev)) {
+		if (!parse_long(o, 0, EV_MAXDEV, &idev)) {
 			return 0;
 		}
-		if (!parse_long(o, EV_MAXDEV, &odev)) {
+		if (!parse_long(o, 0, EV_MAXDEV, &odev)) {
 			return 0;
 		}
 		filt_conf_devmap(f, idev, odev);
@@ -887,7 +1052,7 @@ parse_rule(struct parse_s *o, struct filt_s *f) {
 }
 
 unsigned
-parse_filt(struct parse_s *o, struct filt_s *f) {
+parse_filt(struct parse *o, struct filt *f) {
 	if (!parse_getsym(o)) {
 		return 0;
 	}
@@ -914,8 +1079,8 @@ parse_filt(struct parse_s *o, struct filt_s *f) {
 }
 
 unsigned
-parse_sysex(struct parse_s *o, struct sysex_s **res) {
-	struct sysex_s *sx;
+parse_sysex(struct parse *o, struct sysex **res) {
+	struct sysex *sx;
 	unsigned long val;
 		
 	if (!parse_getsym(o)) {
@@ -944,13 +1109,13 @@ parse_sysex(struct parse_s *o, struct sysex_s **res) {
 						break;
 					}
 					parse_ungetsym(o);
-					if (!parse_long(o, 0xff, &val)) {
+					if (!parse_long(o, 0, 0xff, &val)) {
 						goto err1;
 					}
 					sysex_add(sx, val);
 				}
 			} else if (str_eq(o->lex.strval, "unit")) {
-				if (!parse_long(o, EV_MAXDEV, &val)) {
+				if (!parse_long(o, 0, EV_MAXDEV, &val)) {
 					goto err1;
 				}
 				sx->unit = val;
@@ -978,7 +1143,7 @@ err1:
 }
 
 unsigned
-parse_songchan(struct parse_s *o, struct song_s *s, struct songchan_s *i) {
+parse_songchan(struct parse *o, struct song *s, struct songchan *i) {
 	unsigned long val, val2;
 	if (!parse_getsym(o)) {
 		return 0;
@@ -1003,7 +1168,7 @@ parse_songchan(struct parse_s *o, struct song_s *s, struct songchan_s *i) {
 				if (!parse_nl(o)) {
 					return 0;
 				}
-				track_opsetchan(&i->conf, i->dev, i->ch);
+				track_setchan(&i->conf, i->dev, i->ch);
 			} else if (str_eq(o->lex.strval, "chan")) {
 				if (!parse_chan(o, &val, &val2)) {
 					return 0;
@@ -1039,8 +1204,8 @@ parse_songchan(struct parse_s *o, struct song_s *s, struct songchan_s *i) {
 
 
 unsigned
-parse_songtrk(struct parse_s *o, struct song_s *s, struct songtrk_s *t) {
-	struct songfilt_s *f;
+parse_songtrk(struct parse *o, struct song *s, struct songtrk *t) {
+	struct songfilt *f;
 	unsigned long val;
 	
 	if (!parse_getsym(o)) {
@@ -1069,8 +1234,7 @@ parse_songtrk(struct parse_s *o, struct song_s *s, struct songtrk_s *t) {
 				}
 				f = song_filtlookup(s, o->lex.strval);
 				if (!f) {
-					f = songfilt_new(o->lex.strval);
-					song_filtadd(s, f);
+					f = song_filtnew(s, o->lex.strval);
 				}				
 				t->curfilt = f; 
 				if (!parse_nl(o)) {
@@ -1084,7 +1248,7 @@ parse_songtrk(struct parse_s *o, struct song_s *s, struct songtrk_s *t) {
 					return 0;
 				}
 			} else if (str_eq(o->lex.strval, "mute")) {
-				if (!parse_long(o, 1, &val)) {
+				if (!parse_long(o, 0, 1, &val)) {
 					return 0;
 				}
 				t->mute = val;
@@ -1109,8 +1273,8 @@ parse_songtrk(struct parse_s *o, struct song_s *s, struct songtrk_s *t) {
 
 
 unsigned
-parse_songfilt(struct parse_s *o, struct song_s *s, struct songfilt_s *g) {
-	struct songchan_s *c;
+parse_songfilt(struct parse *o, struct song *s, struct songfilt *g) {
+	struct songchan *c;
 	
 	if (!parse_getsym(o)) {
 		return 0;
@@ -1138,8 +1302,7 @@ parse_songfilt(struct parse_s *o, struct song_s *s, struct songfilt_s *g) {
 				}
 				c = song_chanlookup(s, o->lex.strval);
 				if (!c) {
-					c = songchan_new(o->lex.strval);
-					song_chanadd(s, c);
+					c = song_channew(s, o->lex.strval, 0, 0);
 				}
 				g->curchan = c; 
 				if (!parse_nl(o)) {
@@ -1167,8 +1330,8 @@ parse_songfilt(struct parse_s *o, struct song_s *s, struct songfilt_s *g) {
 
 
 unsigned
-parse_songsx(struct parse_s *o, struct song_s *s, struct songsx_s *g) {
-	struct sysex_s *sx;
+parse_songsx(struct parse *o, struct song *s, struct songsx *g) {
+	struct sysex *sx;
 	
 	if (!parse_getsym(o)) {
 		return 0;
@@ -1206,13 +1369,74 @@ parse_songsx(struct parse_s *o, struct song_s *s, struct songsx_s *g) {
 	return 1;
 }
 
+unsigned
+parse_metro(struct parse *o, struct metro *m) {
+	unsigned long num;
+	struct ev ev;
+
+	if (!parse_getsym(o)) {
+		return 0;
+	}
+	if (o->lex.id != TOK_LBRACE) {
+		lex_err(&o->lex, "'{' expected while parsing metro");
+		return 0;
+	}
+	for (;;) {
+		if (!parse_getsym(o)) {
+			return 0;
+		}
+		if (o->lex.id == TOK_ENDLINE) {
+			/* nothing */
+		} else if (o->lex.id == TOK_RBRACE) {
+			break;
+		} else if (o->lex.id == TOK_IDENT) {
+			if (str_eq(o->lex.strval, "enabled")) {
+				if (!parse_long(o, 0, 1, &num)) {
+					return 0;
+				}
+				if (!parse_nl(o)) {
+					return 0;
+				}
+				m->enabled = num;
+			} else if (str_eq(o->lex.strval, "lo")) {
+				if (!parse_ev(o, &ev)) {
+					return 0;
+				}
+				if (ev.cmd != EV_NON) {
+					lex_err(&o->lex, "'lo' must be followed by a 'non' event\n");
+				} else {
+					m->lo = ev;
+				}					
+			} else if (str_eq(o->lex.strval, "hi")) {
+				if (!parse_ev(o, &ev)) {
+					return 0;
+				}
+				if (ev.cmd != EV_NON) {
+					lex_err(&o->lex, "'hi' must be followed by a 'non' event\n");
+				} else {
+					m->hi = ev;
+				}					
+			} else {
+				goto unknown;
+			}
+		} else {
+		unknown:
+			parse_ungetsym(o);
+			if (!parse_ukline(o)) {
+				return 0;
+			}
+			lex_err(&o->lex, "unknown line format in song, ignored");
+		}
+	}
+	return 1;
+}
 
 unsigned
-parse_song(struct parse_s *o, struct song_s *s) {
-	struct songtrk_s *t;
-	struct songchan_s *i;
-	struct songfilt_s *g;
-	struct songsx_s *l;
+parse_song(struct parse *o, struct song *s) {
+	struct songtrk *t;
+	struct songchan *i;
+	struct songfilt *g;
+	struct songsx *l;
 	unsigned long num, num2;
 
 	if (!parse_getsym(o)) {
@@ -1241,8 +1465,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				}
 				t = song_trklookup(s, o->lex.strval);
 				if (t == NULL) {
-					t = songtrk_new(o->lex.strval);
-					song_trkadd(s, t);
+					t = song_trknew(s, o->lex.strval);
 				}
 				if (!parse_songtrk(o, s, t)) {
 					return 0;
@@ -1260,8 +1483,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				}
 				i = song_chanlookup(s, o->lex.strval);
 				if (i == NULL) {
-					i = songchan_new(o->lex.strval);
-					song_chanadd(s, i);
+					i = song_channew(s, o->lex.strval, 0, 0);
 				}
 				if (!parse_songchan(o, s, i)) {
 					return 0;
@@ -1279,8 +1501,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				}
 				g = song_filtlookup(s, o->lex.strval);
 				if (!g) {
-					g = songfilt_new(o->lex.strval);
-					song_filtadd(s, g);
+					g = song_filtnew(s, o->lex.strval);
 				}
 				if (!parse_songfilt(o, s, g)) {
 					return 0;
@@ -1298,8 +1519,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				}
 				l = song_sxlookup(s, o->lex.strval);
 				if (!l) {
-					l = songsx_new(o->lex.strval);
-					song_sxadd(s, l);
+					l = song_sxnew(s, o->lex.strval);
 				}
 				if (!parse_songsx(o, s, l)) {
 					return 0;
@@ -1315,7 +1535,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 					return 0;
 				}
 			} else if (str_eq(o->lex.strval, "tics_per_unit")) {
-				if (!parse_long(o, ~1U, &num)) {
+				if (!parse_long(o, 0, ~1U, &num)) {
 					return 0;
 				}
 				if ((num % 96) != 0) {
@@ -1329,6 +1549,18 @@ parse_song(struct parse_s *o, struct song_s *s) {
 					return 0;
 				}
 				s->tics_per_unit = num;
+			} else if (str_eq(o->lex.strval, "tempo_factor")) {
+				if (!parse_long(o, 0, ~1U, &num)) {
+					return 0;
+				}
+				if (!parse_nl(o)) {
+					return 0;
+				}
+				if (num < 0x80 || num > 0x200) {
+					lex_err(&o->lex, "warning, tempo factor out of bounds\n");
+				} else {
+					s->tempo_factor = num;
+				}
 			} else if (str_eq(o->lex.strval, "curtrk")) {
 				if (!parse_getsym(o)) {
 					return 0;
@@ -1398,7 +1630,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 					return 0;
 				}
 			} else if (str_eq(o->lex.strval, "curpos")) {
-				if (!parse_long(o, ~1U, &num)) {
+				if (!parse_long(o, 0, ~1U, &num)) {
 					return 0;
 				}
 				if (!parse_nl(o)) {
@@ -1406,7 +1638,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				}
 				s->curpos = num;
 			} else if (str_eq(o->lex.strval, "curlen")) {
-				if (!parse_long(o, ~1U, &num)) {
+				if (!parse_long(o, 0, ~1U, &num)) {
 					return 0;
 				}
 				if (!parse_nl(o)) {
@@ -1414,7 +1646,7 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				}
 				s->curlen = num;
 			} else if (str_eq(o->lex.strval, "curquant")) {
-				if (!parse_long(o, ~1U, &num)) {
+				if (!parse_long(o, 0, ~1U, &num)) {
 					return 0;
 				}
 				if (num > s->tics_per_unit) {
@@ -1434,6 +1666,13 @@ parse_song(struct parse_s *o, struct song_s *s) {
 				if (!parse_nl(o)) {
 					return 0;
 				}
+			} else if (str_eq(o->lex.strval, "metro")) {
+				if (!parse_metro(o, &s->metro)) {
+					return 0;
+				}
+				if (!parse_nl(o)) {
+					return 0;
+				}
 			} else {
 				goto unknown;
 			}
@@ -1450,8 +1689,8 @@ parse_song(struct parse_s *o, struct song_s *s) {
 }
 
 void
-song_save(struct song_s *o, char *name) {
-	struct textout_s *f;
+song_save(struct song *o, char *name) {
+	struct textout *f;
 	
 	f = textout_new(name);
 	if (f == NULL) {
@@ -1468,8 +1707,8 @@ song_save(struct song_s *o, char *name) {
 }
 
 unsigned
-song_load(struct song_s *o, char *filename) {
-	struct parse_s *parse;
+song_load(struct song *o, char *filename) {
+	struct parse *parse;
 	unsigned res;
 	
 	parse = parse_new(filename);
