@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003-2005 Alexandre Ratchov
+ * Copyright (c) 2003-2007 Alexandre Ratchov <alex@caoua.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -36,10 +36,13 @@
 #include "dbg.h"
 #include "sysex.h"
 #include "track.h"
-#include "trackop.h"
 #include "song.h"
 #include "smf.h"
 #include "cons.h"
+#include "frame.h"
+#include "conv.h"
+
+#define MAXTRACKNAME 100
 
 char smftype_header[4] = { 'M', 'T', 'h', 'd' };
 char smftype_track[4]  = { 'M', 'T', 'r', 'k' };
@@ -49,18 +52,17 @@ unsigned smf_evlen[] = { 2, 2, 2, 2, 1, 1, 2, 0 };
 
 /* --------------------------------------------- chunk read/write --- */
 
-struct smf_s {
+struct smf {
 	FILE *file;
 	unsigned length, index;		/* current chunk length/position */
 };
 
-	/*
-	 * open a standard midi file and initialise
-	 * the smf_s structure
-	 */
-
+/*
+ * open a standard midi file and initialize
+ * the smf structure
+ */
 unsigned
-smf_open(struct smf_s *o, char *path, char *mode) {
+smf_open(struct smf *o, char *path, char *mode) {
 	o->file = fopen(path, mode);
 	if (!o->file) {
 		cons_errs(path, "failed to open file");
@@ -71,13 +73,19 @@ smf_open(struct smf_s *o, char *path, char *mode) {
 	return 1;	
 }
 
+/*
+ * close the file
+ */
 void
-smf_close(struct smf_s *o) {
+smf_close(struct smf *o) {
 	fclose(o->file);
 }
 
+/*
+ * read a 32bit fixed-size number, return 0 on error
+ */
 unsigned
-smf_get32(struct smf_s *o, unsigned *val) {
+smf_get32(struct smf *o, unsigned *val) {
 	unsigned char buf[4];
 	if (o->index + 4 > o->length || fread(buf, 1, 4, o->file) != 4) {
 		cons_err("failed to read 32bit number");
@@ -88,8 +96,11 @@ smf_get32(struct smf_s *o, unsigned *val) {
 	return 1;
 }
 
+/*
+ * read a 24bit fixed-size number, return 0 on error
+ */
 unsigned
-smf_get24(struct smf_s *o, unsigned *val) {
+smf_get24(struct smf *o, unsigned *val) {
 	unsigned char buf[4];
 	if (o->index + 3 > o->length || fread(buf, 1, 3, o->file) != 3) {
 		cons_err("failed to read 24bit number");
@@ -100,8 +111,11 @@ smf_get24(struct smf_s *o, unsigned *val) {
 	return 1;
 }
 
+/*
+ * read a 16bit fixed-size number, return 0 on error
+ */
 unsigned
-smf_get16(struct smf_s *o, unsigned *val) {
+smf_get16(struct smf *o, unsigned *val) {
 	unsigned char buf[4];
 	if (o->index + 2 > o->length || fread(buf, 1, 2, o->file) != 2) {
 		cons_err("failed to read 16bit number");
@@ -112,8 +126,11 @@ smf_get16(struct smf_s *o, unsigned *val) {
 	return 1;
 }
 
+/*
+ * read a 8bit fixed-size number, return 0 on error
+ */
 unsigned
-smf_getc(struct smf_s *o, unsigned *res) {
+smf_getc(struct smf *o, unsigned *res) {
 	int c;
 	if (o->index + 1 > o->length || (c = fgetc(o->file)) < 0) {
 		cons_err("failed to read one byte");
@@ -124,8 +141,11 @@ smf_getc(struct smf_s *o, unsigned *res) {
 	return 1;
 }
 
+/*
+ * read a variable length number, return 0 on error
+ */
 unsigned
-smf_getvar(struct smf_s *o, unsigned *val) {
+smf_getvar(struct smf *o, unsigned *val) {
 	int c;
 	unsigned bits;
 	*val = 0;
@@ -153,9 +173,14 @@ smf_getvar(struct smf_s *o, unsigned *val) {
 	return 1;
 }
 
-		
+
+/*
+ * read a chunk header, compare it with ethe given 4-byte header and
+ * initialize the smf structure so that other smf_getxxx can work.
+ * return 0 on error
+ */
 unsigned
-smf_getheader(struct smf_s *o, char *hdr) {
+smf_getheader(struct smf *o, char *hdr) {
 	char buf[4];
 	unsigned len;
 	if (o->index != o->length) {
@@ -180,17 +205,19 @@ smf_getheader(struct smf_s *o, char *hdr) {
 	return 1;
 }
 
-	/*
-	 * in smf_put* routines there is a 'unsigned *used' argument
-	 * if (used == NULL) then data is not written in the smf_s
-	 * structure, only *user is incremented by the number
-	 * of bytes that would be written otherwise.
-	 * This is used to determine chunk's length
-	 */
+/*
+ * in smf_putxxx routines there is a 'unsigned *used' argument if
+ * (used == NULL) then data is not written in the smf structure, only
+ * *user is incremented by the number of bytes that would be written
+ * otherwise.  This is used to determine chunk's length
+ */
 
 
+/*
+ * put a fixed-size 32-bit number
+ */
 void
-smf_put32(struct smf_s *o, unsigned *used, unsigned val) {
+smf_put32(struct smf *o, unsigned *used, unsigned val) {
 	unsigned char buf[4];
 	if (used) {
 		(*used) += 4;
@@ -208,8 +235,11 @@ smf_put32(struct smf_s *o, unsigned *used, unsigned val) {
 	}
 }
 
+/*
+ * put a fixed-size 24-bit number
+ */
 void
-smf_put24(struct smf_s *o, unsigned *used, unsigned val) {
+smf_put24(struct smf *o, unsigned *used, unsigned val) {
 	unsigned char buf[4];
 	if (used) {
 		(*used) += 3;
@@ -226,8 +256,11 @@ smf_put24(struct smf_s *o, unsigned *used, unsigned val) {
 	}
 }
 
+/*
+ * put a fixed-size 16-bit number
+ */
 void
-smf_put16(struct smf_s *o, unsigned *used, unsigned val) {
+smf_put16(struct smf *o, unsigned *used, unsigned val) {
 	unsigned char buf[4];
 	if (used) {
 		(*used) += 2;
@@ -244,8 +277,11 @@ smf_put16(struct smf_s *o, unsigned *used, unsigned val) {
 }
 
 
+/*
+ * put a fixed-size 8-bit number
+ */
 void
-smf_putc(struct smf_s *o, unsigned *used, unsigned val) {
+smf_putc(struct smf *o, unsigned *used, unsigned val) {
 	if (used) {
 		(*used) ++;
 	} else {
@@ -259,9 +295,12 @@ smf_putc(struct smf_s *o, unsigned *used, unsigned val) {
 }
 
 
+/*
+ * put a variable length number
+ */
 void
-smf_putvar(struct smf_s *o, unsigned *used, unsigned val) {
-#define MAXBYTES (5)			/* 32bit / 7bit = 4bytes + 4bit */
+smf_putvar(struct smf *o, unsigned *used, unsigned val) {
+#define MAXBYTES 5			/* 32bit / 7bit = 4bytes + 4bit */
 	unsigned char buf[MAXBYTES];
 	unsigned index = 0, bits;
 	for (bits = 7; bits < MAXBYTES * 7; bits += 7) {
@@ -289,8 +328,11 @@ smf_putvar(struct smf_s *o, unsigned *used, unsigned val) {
 #undef MAXBYTES
 }
 
+/*
+ * put the given chunk header with the given magic and size field
+ */
 void
-smf_putheader(struct smf_s *o, char *hdr, unsigned len) {
+smf_putheader(struct smf *o, char *hdr, unsigned len) {
 	fwrite(hdr, 1, 4, o->file);
 	o->length = 4;
 	o->index = 0;
@@ -299,29 +341,56 @@ smf_putheader(struct smf_s *o, char *hdr, unsigned len) {
 	o->index = 0;
 }
 
-/* ------------------------------------------------------- export --- */
-
+/*
+ * store a track in the smf
+ */
 void
-smf_putmeta(struct smf_s *o, unsigned *used, struct song_s *s) {
-	struct seqptr_s tp;
-	unsigned delta, denom;
+smf_puttrack(struct smf *o, unsigned *used, struct song *s, struct track *t) {
+	struct seqev *pos;
+	unsigned status, newstatus, delta, chan, denom;
+	struct ev rev[CONV_NUMREV];
+	struct statelist slist;
+	unsigned i, nev;
 	
+
+	statelist_init(&slist);
 	delta = 0;
-	track_rew(&s->meta, &tp);
-	for (;;) {
-		delta += track_ticlast(&s->meta, &tp);
-		if (!track_evavail(&s->meta, &tp)) {
+	status = 0;
+	for (pos = t->first; pos != NULL; pos = pos->next) {
+		delta += pos->delta;
+		if (pos->ev.cmd == EV_NULL) {
 			break;
 		}
-		if ((*tp.pos)->ev.cmd == EV_TEMPO) {
+		if (EV_ISVOICE(&pos->ev)) {
+			nev = conv_unpackev(&slist, ~1U, &pos->ev, rev); 
+			for (i = 0; i < nev; i++) {
+				smf_putvar(o, used, delta);
+				delta = 0;
+				chan = rev[i].ch;
+				newstatus = (rev[i].cmd << 4) + (chan & 0x0f);
+				if (newstatus != status) {
+					status = newstatus;
+					smf_putc(o, used, status);
+				}
+				if (rev[i].cmd == EV_BEND) {
+					smf_putc(o, used, rev[i].bend_val & 0x7f);
+					smf_putc(o, used, rev[i].bend_val >> 7);
+				} else {
+					smf_putc(o, used, rev[i].v0);
+					if (SMF_EVLEN(status) == 2) {
+						smf_putc(o, used, rev[i].v1);
+					}
+				}
+			}
+		} else if (pos->ev.cmd == EV_TEMPO) {
 			smf_putvar(o, used, delta);
 			delta = 0;
 			smf_putc(o, used, 0xff);
 			smf_putc(o, used, 0x51);
 			smf_putc(o, used, 0x03);
-			smf_put24(o, used, (*tp.pos)->ev.data.tempo.usec24 * s->tics_per_unit / 96);
-		} else if ((*tp.pos)->ev.cmd == EV_TIMESIG) {
-			denom = s->tics_per_unit / (*tp.pos)->ev.data.sign.tics;
+			smf_put24(o, used, pos->ev.tempo_usec24 * s->tics_per_unit / 96);
+		} else if (pos->ev.cmd == EV_TIMESIG) {
+			denom = s->tics_per_unit / pos->ev.timesig_tics;
 			switch(denom) {
 			case 1:	
 				denom = 0; 
@@ -339,7 +408,7 @@ smf_putmeta(struct smf_s *o, unsigned *used, struct song_s *s) {
 				denom = 4;
 				break;
 			default:
-				dbg_puts("smf_putmeta: bad time signature\n");
+				dbg_puts("smf_puttrack: bad time signature\n");
 				dbg_panic();
 			}
 			smf_putvar(o, used, delta);
@@ -347,99 +416,28 @@ smf_putmeta(struct smf_s *o, unsigned *used, struct song_s *s) {
 			smf_putc(o, used, 0xff);
 			smf_putc(o, used, 0x58);
 			smf_putc(o, used, 0x04);
-			smf_putc(o, used, (*tp.pos)->ev.data.sign.beats);
+			smf_putc(o, used, pos->ev.timesig_beats);
 			smf_putc(o, used, denom);
 			/* metronome tics per metro beat */
-			smf_putc(o, used, (*tp.pos)->ev.data.sign.tics);
+			smf_putc(o, used, pos->ev.timesig_tics);
 			/* metronome 1/32 notes per 24 tics */
 			smf_putc(o, used, 8 * s->tics_per_unit / 96);
 		}
 
-		track_evnext(&s->meta, &tp);
 	}
 	smf_putvar(o, used, delta);
 	smf_putc(o, used, 0xff);
 	smf_putc(o, used, 0x2f);
 	smf_putc(o, used, 0x00);
-	return;
+	statelist_done(&slist);
 }
 
+/*
+ * store a sysex in the smf
+ */
 void
-smf_puttrk(struct smf_s *o, unsigned *used, struct song_s *s, struct songtrk_s *t) {
-	struct seqptr_s tp;
-	unsigned status, newstatus, delta, chan;
-	
-	status = 0;
-	delta = 0;
-	track_rew(&t->track, &tp);
-	for (;;) {
-		delta += track_ticlast(&t->track, &tp);
-		if (!track_evavail(&t->track, &tp)) {
-			break;
-		}
-		if (EV_ISVOICE(&(*tp.pos)->ev)) {
-			smf_putvar(o, used, delta);
-			delta = 0;
-			chan = (*tp.pos)->ev.data.voice.ch;
-			newstatus = ((*tp.pos)->ev.cmd << 4) + (chan & 0x0f);
-			if (newstatus != status) {
-				status = newstatus;
-				smf_putc(o, used, status);
-			}
-			smf_putc(o, used, (*tp.pos)->ev.data.voice.b0);
-			if (SMF_EVLEN(status) == 2) {
-				smf_putc(o, used, (*tp.pos)->ev.data.voice.b1);
-			}
-		}
-		track_evnext(&t->track, &tp);
-	}
-	smf_putvar(o, used, delta);
-	smf_putc(o, used, 0xff);
-	smf_putc(o, used, 0x2f);
-	smf_putc(o, used, 0x00);
-	return;
-}
-
-
-void
-smf_putchan(struct smf_s *o, unsigned *used, struct song_s *s, struct songchan_s *i) {
-	struct seqptr_s tp;
-	unsigned status, newstatus, delta;
-	
-	status = 0;
-	delta = 0;
-	track_rew(&i->conf, &tp);
-	for (;;) {
-		delta += track_ticlast(&i->conf, &tp);
-		if (!track_evavail(&i->conf, &tp)) {
-			break;
-		}
-		if (EV_ISVOICE(&(*tp.pos)->ev)) {
-			smf_putvar(o, used, delta);
-			delta = 0;
-			newstatus = ((*tp.pos)->ev.cmd << 4) + (i->ch & 0x0f);
-			if (newstatus != status) {
-				status = newstatus;
-				smf_putc(o, used, status);
-			}
-			smf_putc(o, used, (*tp.pos)->ev.data.voice.b0);
-			if (SMF_EVLEN(status) == 2) {
-				smf_putc(o, used, (*tp.pos)->ev.data.voice.b1);
-			}
-		}
-		track_evnext(&i->conf, &tp);
-	}
-	smf_putvar(o, used, delta);
-	smf_putc(o, used, 0xff);
-	smf_putc(o, used, 0x2f);
-	smf_putc(o, used, 0x00);
-	return;
-}
-
-
-void
-smf_putsysex(struct smf_s *o, unsigned *used, struct sysex_s *sx) {
-	struct chunk_s *c;
+smf_putsysex(struct smf *o, unsigned *used, struct sysex *sx) {
+	struct chunk *c;
 	unsigned i, first;
 		
 	first = 1;
@@ -454,10 +452,13 @@ smf_putsysex(struct smf_s *o, unsigned *used, struct sysex_s *sx) {
 	}
 }
 
+/*
+ * store a sysex back in the smf
+ */
 void
-smf_putsx(struct smf_s *o, unsigned *used, struct song_s *s, struct songsx_s *songsx) {
+smf_putsx(struct smf *o, unsigned *used, struct song *s, struct songsx *songsx) {
 	unsigned sysexused;
-	struct sysex_s *sx;
+	struct sysex *sx;
 	
 	for (sx = songsx->sx.first; sx != NULL; sx = sx->next) {
 		sysexused = 0;
@@ -473,78 +474,87 @@ smf_putsx(struct smf_s *o, unsigned *used, struct song_s *s, struct songsx_s *so
 	smf_putc(o, used, 0x00);		
 }
 
+/*
+ * open the smf and store the whole song
+ */
 unsigned
-song_exportsmf(struct song_s *o, char *filename) {
-	struct smf_s f;
-	struct songtrk_s *t;
-	struct songchan_s *i;
-	struct songsx_s *s;
+song_exportsmf(struct song *o, char *filename) {
+	struct smf f;
+	struct songtrk *t;
+	struct songchan *i;
+	struct songsx *s;
 	unsigned ntrks, nchan, nsx, used;
 
 	if (!smf_open(&f, filename, "w")) {
 		return 0;
 	}
 	ntrks = 0;
-	for (t = o->trklist; t != NULL; t = (struct songtrk_s *)t->name.next) {
+	SONG_FOREACH_TRK(o, t) {
 		ntrks++;
 	}
 	nchan = 0;
-	for (i = o->chanlist; i != NULL; i = (struct songchan_s *)i->name.next) {
+	SONG_FOREACH_CHAN(o, i) {
 		nchan++;
 	}
 	nsx = 0;
-	for (s = o->sxlist; s != NULL; s = (struct songsx_s *)s->name.next) {
+	SONG_FOREACH_SX(o, s) {
 		nsx++;
 	}
 
-	/* write the header */
+	/* 
+	 * write the header 
+	 */
 	smf_putheader(&f, smftype_header, 6);
 	smf_put16(&f, NULL, 1);				/* format = 1 */
 	smf_put16(&f, NULL, nsx + ntrks + nchan + 1);	/* +1 -> meta track */
 	smf_put16(&f, NULL, o->tics_per_unit / 4);		/* tics per quarter */
 
-	/* write the tempo track */
+	/* 
+	 * write the tempo track 
+	 */
 	used = 0;
-	smf_putmeta(&f, &used, o);
+	smf_puttrack(&f, &used, o, &o->meta);
 	smf_putheader(&f, smftype_track, used);
-	smf_putmeta(&f, NULL, o);
+	smf_puttrack(&f, NULL, o, &o->meta);
 		
-	/* write each sx */
-	for (s = o->sxlist; s != NULL; s = (struct songsx_s *)s->name.next) {
+	/* 
+	 * write each sx 
+	 */
+	SONG_FOREACH_SX(o, s) {
 		used = 0;
 		smf_putsx(&f, &used, o, s);
 		smf_putheader(&f, smftype_track, used);
 		smf_putsx(&f, NULL, o, s);
 	}	
 					
-	/* write each chan */
-	for (i = o->chanlist; i != NULL; i = (struct songchan_s *)i->name.next) {
+	/* 
+	 * write each chan 
+	 */
+	SONG_FOREACH_CHAN(o, i) {
 		used = 0;
-		smf_putchan(&f, &used, o, i);
+		smf_puttrack(&f, &used, o, &i->conf);
 		smf_putheader(&f, smftype_track, used);
-		smf_putchan(&f, NULL, o, i);
+		smf_puttrack(&f, NULL, o, &i->conf);
 	}
 
-	/* write each track */
-	for (t = o->trklist; t != NULL; t = (struct songtrk_s *)t->name.next) {
+	/*
+	 * write each track 
+	 */
+	SONG_FOREACH_TRK(o, t) {
 		used = 0;
-		smf_puttrk(&f, &used, o, t);
+		smf_puttrack(&f, &used, o, &t->track);
 		smf_putheader(&f, smftype_track, used);
-		smf_puttrk(&f, NULL, o, t);
+		smf_puttrack(&f, NULL, o, &t->track);
 	}
 	smf_close(&f);
 	return 1;
 }
 
-/* ------------------------------------------------------- import --- */
-
-
-	/* 
-	 * parse '0xF0 varlen data data ... data 0xF7'
-	 */
-
+/* 
+ * parse '0xF0 varlen data data ... data 0xF7'
+ */
 unsigned
-smf_getsysex(struct smf_s *o, struct sysex_s *sx) {
+smf_getsysex(struct smf *o, struct sysex *sx) {
 	unsigned i, length, c;
 	
 	if (!smf_getvar(o, &length)) {
@@ -559,18 +569,18 @@ smf_getsysex(struct smf_s *o, struct sysex_s *sx) {
 	return 1;
 }
 
-
-	/*
-	 * parse a track 'varlen event varlen event ... varlen event'
-	 */
-
+/*
+ * parse a track 'varlen event varlen event ... varlen event'
+ */
 unsigned
-smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
+smf_gettrack(struct smf *o, struct song *s, struct songtrk *t) {
 	unsigned delta, i, status, type, length, abspos;
 	unsigned tempo, num, den, dummy;
-	struct seqptr_s tp;
-	struct ev_s ev;
-	struct sysex_s *sx;
+	struct statelist slist;
+	struct songsx *songsx;
+	struct seqev *pos, *se;
+	struct sysex *sx;
+	struct ev ev, rev;
 	unsigned c;
 	
 	if (!smf_getheader(o, smftype_track)) {
@@ -578,27 +588,31 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 	}
 	status = 0;
 	abspos = 0;
-	track_clear(&t->track, &tp);
+	track_clear(&t->track);
+	pos = t->track.first;
+	songsx = (struct songsx *)s->sxlist;	/* first (and unique) sysex in song */
+	statelist_init(&slist);
 	for (;;) {
 		if (o->index >= o->length) {
+			statelist_done(&slist);
 			return 1;
 		}
 		if (!smf_getvar(o, &delta)) {
-			return 0;
+			goto err;
 		}
 		abspos += delta;
-		track_seekblank(&t->track, &tp, delta);
+		pos->delta += delta;
 		if (!smf_getc(o, &c)) {
-			return 0;
+			goto err;;
 		}
 		if (c == 0xff) {
 			status = 0;
 			if (!smf_getc(o, &c)) {
-				return 0;
+				goto err;
 			}
 			type = c;
 			if (!smf_getvar(o, &length)) {
-				return 0;
+				goto err;
 			}
 			if (type == 0x2f && length == 0) {
 				/* end of track */
@@ -606,24 +620,24 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 			} else if (type == 0x51 && length == 3) {
 				/* tempo change */
 				if (!smf_get24(o, &tempo)) {
-					return 0;
+					goto err;
 				}
 				ev.cmd = EV_TEMPO;
-				ev.data.tempo.usec24 = tempo * 96 / s->tics_per_unit;
+				ev.tempo_usec24 = tempo * 96 / s->tics_per_unit;
 				goto putev;
 			} else if (type == 0x58 && length == 4) {
 				/* time signature change */
 				if (!smf_getc(o, &num)) {
-					return 0;
+					goto err;
 				}
 				if (!smf_getc(o, &den)) {
-					return 0;
+					goto err;
 				}
 				ev.cmd = EV_TIMESIG;
-				ev.data.sign.beats = num;
-				ev.data.sign.tics = s->tics_per_unit / (1 << den);
+				ev.timesig_beats = num;
+				ev.timesig_tics = s->tics_per_unit / (1 << den);
 				if (!smf_getc(o, &dummy)) {
-					return 0;
+					goto err;
 				}
 				/*
 				dbg_puts("cc=");
@@ -631,7 +645,7 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 				dbg_puts(", dd=");
 				*/
 				if (!smf_getc(o, &dummy)) {
-					return 0;
+					goto err;
 				}
 				/*
 				dbg_putu(dummy);
@@ -642,7 +656,7 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 			ignoremeta:
 				for (i = 0; i < length; i++) {
 					if (!smf_getc(o, &c)) {
-						return 0;
+						goto err;
 					}
 				}
 			}
@@ -650,27 +664,25 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 			/* raw data */
 			cons_err("raw data (status = 0xF7) not implemented");
 			status = 0;
-			ev.cmd = 0;
 			if (!smf_getvar(o, &length)) {
-				return 0;
+				goto err;
 			}
 			for (i = 0; i < length; i++) {
 				if (!smf_getc(o, &c)) {
-					return 0;
+					goto err;
 				}
 			}
 		} else if (c == 0xf0) {
 			/* sys ex */
 			status = 0;
-			ev.cmd = 0;
 			sx = sysex_new(0);
 			sysex_add(sx, 0xf0);
 			if (!smf_getsysex(o, sx)) {
 				sysex_del(sx);
-				return 0;
+				goto err;
 			}				
 			if (sysex_check(sx)) {
-				sysexlist_put(&s->sxlist->sx, sx);
+				sysexlist_put(&songsx->sx, sx);
 			} else {
 				cons_err("corrupted sysex message, ignored");
 				/*
@@ -682,25 +694,34 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 		} else if (c >= 0x80 && c < 0xf0) {
 			status = c;
 			if (!smf_getc(o, &c)) {
-				return 0;
+				goto err;
 			}
 		runningstatus:
 			ev.cmd = (status >> 4) & 0xf;
-			ev.data.voice.dev = 0;
-			ev.data.voice.ch = status & 0xf;
-			ev.data.voice.b0 = c & 0x7f;
+			ev.dev = 0;
+			ev.ch = status & 0xf;
+			ev.v0 = c & 0x7f;
 			if (SMF_EVLEN(status) == 2) {
 				if (!smf_getc(o, &c)) {
-					return 0;
+					goto err;
+				}
+				c &= 0x7f;
+				if (ev.cmd == EV_BEND) {
+					ev.v0 += c << 7;
+				} else {
+					ev.v1 = c;
 				}
-				ev.data.voice.b1 = c & 0x7f;
 			}
 		putev:
-			if (ev.cmd == EV_NON && ev.data.voice.b1 == 0) {
+			if (ev.cmd == EV_NON && ev.note_vel == 0) {
 				ev.cmd = EV_NOFF;
-				ev.data.voice.b1 = EV_NOFF_DEFAULTVEL;
+				ev.note_vel = EV_NOFF_DEFAULTVEL;
+			}
+			if (conv_packev(&slist, 0U, &ev, &rev)) {
+				se = seqev_new();
+				se->ev = rev;
+				seqev_ins(pos, se);
 			}
-			track_evput(&t->track, &tp, &ev);
 			/*
 			dbg_puts("ev: ");
 			dbg_putu(abspos);
@@ -711,78 +732,140 @@ smf_gettrack(struct smf_s *o, struct song_s *s, struct songtrk_s *t) {
 		} else if (c < 0x80) {
 			if (status == 0) {
 				cons_err("bad status");
-				return 0;
+				goto err;
 			}
 			goto runningstatus;
 		} else {
 			cons_err("bad event");
-			return 0;
+			goto err;
 		}
 	}
-	return 1;
+ err:
+	statelist_done(&slist);
+	return 0;
 }
 
-	/*
-	 * fix song imported from format 1 smf 
-	 */
-
+/*
+ * fix song imported from format 1 smf 
+ */
 void
-song_fix1(struct song_s *o) {
-	struct seqptr_s mp, tp;
-	struct seqev_s *se;
-	struct songtrk_s *t;
+song_fix1(struct song *o) {
+	struct track copy;
+	struct seqptr tp, cp;
+	struct state *st;
+	struct statelist slist;
+	struct songtrk *t, *tnext;
 	unsigned delta;
 		
-	for (t = o->trklist; t != NULL; t = (struct songtrk_s *)t->name.next) {
-		/* move meta events into meta track */
+	SONG_FOREACH_TRK(o, t) {
+		/* 
+		 * move meta events into meta track 
+		 */
 		delta = 0;
-		track_rew(&o->meta, &mp);
-		track_rew(&t->track, &tp);
+		track_init(&copy);
+		seqptr_init(&cp, &copy);
+		seqptr_init(&tp, &t->track);
+		statelist_init(&slist);
 		for (;;) {
-			delta = track_ticlast(&t->track, &tp);
-			track_seekblank(&o->meta, &mp, delta);
-			track_evlast(&o->meta, &mp);
-			
-			if (!track_evavail(&t->track, &tp)) {
+			delta = seqptr_ticdel(&tp, ~0U, &slist);
+			seqptr_ticput(&tp, delta);
+			seqptr_ticput(&cp, delta);
+			st = seqptr_evdel(&tp, &slist);
+			if (st == NULL) {
 				break;
 			}
-			if (EV_ISMETA(&(*tp.pos)->ev)) {
-				se = track_seqevrm(&t->track, &tp);
-				track_seqevins(&o->meta, &mp, se);
-				track_seqevnext(&o->meta, &mp);
+			if (st->phase & EV_PHASE_FIRST) {
+				st->tag = EV_ISMETA(&st->ev) ? 1 : 0;
+			}
+			if (st->tag) {
+				seqptr_evput(&cp, &st->ev);
 			} else {
-				track_evnext(&t->track, &tp);
-			}				
+				seqptr_evput(&tp, &st->ev);
+			}
+		}
+		statelist_done(&slist);
+		seqptr_done(&tp);
+		seqptr_done(&cp);
+		track_merge(&o->meta, &copy);
+		track_done(&copy);
+	}
+
+	/*
+	 * remove the first track, if there are not events
+	 */
+	for (t = (struct songtrk *)o->trklist; t != NULL; t = tnext) {
+		tnext = (struct songtrk *)t->name.next;
+		if (t->track.first->ev.cmd == EV_NULL) {
+			song_trkdel(o, t);
 		}
-		/* check for inconsistecies */
-		track_opcheck(&t->track);
 	}
 }
 
-	/*
-	 * fix song imported from format 0 smf 
-	 */
-
+/*
+ * fix song imported from format 0 SMFs: call fix1 to move meta events
+ * to the meta-track then split it creating one track per channel
+ */
 void
-song_fix0(struct song_s *o) {
+song_fix0(struct song *o) {
+	char trackname[MAXTRACKNAME];	
+	struct seqptr tp, cp;
+	struct state *st;
+	struct statelist slist;
+	struct songtrk *t, *smf;
+	unsigned delta;
+	unsigned i;
+
 	song_fix1(o);
+	smf = (struct songtrk *)o->trklist;
+	if (smf == NULL) {
+		return;
+	}
+	for (i = 1; i < 16; i++) {
+		snprintf(trackname, MAXTRACKNAME, "trk%02u", i);
+		t = song_trknew(o, trackname);
+		delta = 0;
+		seqptr_init(&cp, &t->track);
+		seqptr_init(&tp, &smf->track);
+		statelist_init(&slist);
+		for (;;) {
+			delta = seqptr_ticdel(&tp, ~0U, &slist);
+			seqptr_ticput(&tp, delta);
+			seqptr_ticput(&cp, delta);
+			st = seqptr_evdel(&tp, &slist);
+			if (st == NULL) {
+				break;
+			}
+			if (st->phase & EV_PHASE_FIRST) {
+				if (EV_ISVOICE(&st->ev) &&
+				    st->ev.ch == i) {
+					st->tag = 1;
+				} else {
+					st->tag = 0;
+				}
+			}
+			if (st->tag) {
+				seqptr_evput(&cp, &st->ev);
+			} else {
+				seqptr_evput(&tp, &st->ev);
+			}
+		}
+		statelist_done(&slist);
+		seqptr_done(&tp);
+		seqptr_done(&cp);
+	}
 }
 
-#define MAXTRACKNAME 100
-
-struct song_s *
+/*
+ * open the smf and load the whole file
+ */
+struct song *
 song_importsmf(char *filename) {
-	struct song_s *o;
-	struct songtrk_s *t;
+	struct song *o;
+	struct songtrk *t;
 	unsigned format, ntrks, timecode, i;
 	char trackname[MAXTRACKNAME];	
-	struct songsx_s *sx;
-	struct smf_s f;
-	
-	/*
-	cons_err("not implemented");
-	return 0;
-	*/
+	struct songsx *sx;
+	struct smf f;
 	
 	if (!smf_open(&f, filename, "r")) {
 		goto bad1;
@@ -810,17 +893,16 @@ song_importsmf(char *filename) {
 	
 	o = song_new();
 	o->tics_per_unit = timecode * 4;   /* timecode = tics per quarter */
-	sx = songsx_new("sx");
-	song_sxadd(o, sx);
+	sx = song_sxnew(o, "sx");
 	o->cursx = sx;
 	
 	for (i = 0; i < ntrks; i++) {
 		snprintf(trackname, MAXTRACKNAME, "trk%02u", i);
-		t = songtrk_new(trackname);
-		song_trkadd(o, t);
+		t = song_trknew(o, trackname);
 		if (!smf_gettrack(&f, o, t)) {
 			goto bad3;
 		}
+		track_check(&t->track);
 	}
 	smf_close(&f);
 	
@@ -831,8 +913,9 @@ song_importsmf(char *filename) {
 	}
 	
 	
-	/* TODO: move sysex messages into separate songsx */
-	
+	/* 
+	 * TODO: move sysex messages into separate songsx 
+	 */	
 	return o;
 	
 bad3:	song_delete(o);
