@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003-2005 Alexandre Ratchov
+ * Copyright (c) 2003-2007 Alexandre Ratchov <alex@caoua.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -29,8 +29,10 @@
  */
 
 /*
- * implements misc built-in functions
- * available through the interpreter
+ * implements misc built-in functions available through the
+ * interpreter
+ *
+ * each function is described in the manual.html file
  */
 
 #include "dbg.h"
@@ -47,7 +49,6 @@
 #include "mux.h"
 #include "mididev.h"
 
-#include "trackop.h"
 #include "track.h"
 #include "song.h"
 #include "user.h"
@@ -55,25 +56,24 @@
 #include "saveload.h"
 #include "rmidi.h"	/* for rmidi_debug */
 
-struct song_s *user_song;
+struct song *user_song;
 unsigned user_flag_batch = 0;
 unsigned user_flag_verb = 0;
 
 /* -------------------------------------------------- some tools --- */
 
-	/*
-	 * execute a script from a file in the 'exec' environnement
-	 * the script has acces to the global variables, but
-	 * not to the local variables of the calling proc. Thus
-	 * it can be safely used in a procedure
-	 */
-
+/*
+ * execute the script in the given file inside the 'exec' environment.
+ * the script has acces to all global variables, but not to the local
+ * variables of the calling proc. Thus it can be safely used from a
+ * procedure
+ */
 unsigned
-exec_runfile(struct exec_s *exec, char *filename) {
-	struct parse_s *parse;
-	struct var_s **locals;
-	struct node_s *root;
-	struct data_s *data;
+exec_runfile(struct exec *exec, char *filename) {
+	struct parse *parse;
+	struct name **locals;
+	struct node *root;
+	struct data *data;
 	unsigned res;
 	
 	res = 0;
@@ -86,7 +86,6 @@ exec_runfile(struct exec_s *exec, char *filename) {
 	locals = exec->locals;
 	exec->locals = &exec->globals;
 	if (parse_prog(parse, &root)) {
-		/*node_dbg(root, 0);*/
 		res = node_exec(root, exec, &data);
 	}
 	exec->locals = locals;
@@ -95,15 +94,14 @@ exec_runfile(struct exec_s *exec, char *filename) {
 	return res;
 }
 
-	/*
-	 * find the pointer to the songtrk contained in 'var'
-	 * ('var' must be a reference)
-	 */
-
+/*
+ * find the pointer to the songtrk contained in 'var' ('var' must be a
+ * reference)
+ */
 unsigned
-exec_lookuptrack(struct exec_s *o, char *var, struct songtrk_s **res) {
+exec_lookuptrack(struct exec *o, char *var, struct songtrk **res) {
 	char *name;	
-	struct songtrk_s *t;
+	struct songtrk *t;
 	if (!exec_lookupname(o, var, &name)) {
 		return 0;
 	}
@@ -116,18 +114,16 @@ exec_lookuptrack(struct exec_s *o, char *var, struct songtrk_s **res) {
 	return 1;
 }
 
-
-	/*
-	 * find the (dev, ch) couple for the channel referenced by 'var'
-	 * 'var' can be 
-	 *	- a reference to a songchan
-	 *	- a list of two integers (like '{dev chan}')
-	 */
-
+/*
+ * find the (dev, ch) couple for the channel referenced by 'var'
+ * 'var' can be 
+ *	- a reference to a songchan
+ *	- a list of two integers (like '{dev chan}')
+ */
 unsigned
-exec_lookupchan_getnum(struct exec_s *o, char *var, 
+exec_lookupchan_getnum(struct exec *o, char *var, 
     unsigned *dev, unsigned *ch) {
-	struct var_s *arg;
+	struct var *arg;
 	
 	arg = exec_varlookup(o, var);
 	if (!arg) {
@@ -140,16 +136,14 @@ exec_lookupchan_getnum(struct exec_s *o, char *var,
 	return 1;
 }
 
-	/*
-	 * find the pointer to an existing songchan
-	 * that is referenced by 'var'.
-	 * ('var' must be a reference)
-	 */
-
+/*
+ * find the pointer to an existing songchan that is referenced by
+ * 'var'.  ('var' must be a reference)
+ */
 unsigned
-exec_lookupchan_getref(struct exec_s *o, char *var, struct songchan_s **res) {
-	struct var_s *arg;
-	struct songchan_s *i;
+exec_lookupchan_getref(struct exec *o, char *var, struct songchan **res) {
+	struct var *arg;
+	struct songchan *i;
 	
 	arg = exec_varlookup(o, var);
 	if (!arg) {
@@ -170,16 +164,14 @@ exec_lookupchan_getref(struct exec_s *o, char *var, struct songchan_s **res) {
 	return 1;
 }
 
-
-	/*
-	 * find the pointer to the songfilt contained in 'var'
-	 * ('var' must be a reference)
-	 */
-
+/*
+ * find the pointer to the songfilt contained in 'var' ('var' must be
+ * a reference)
+ */
 unsigned
-exec_lookupfilt(struct exec_s *o, char *var, struct songfilt_s **res) {
+exec_lookupfilt(struct exec *o, char *var, struct songfilt **res) {
 	char *name;	
-	struct songfilt_s *f;
+	struct songfilt *f;
 	if (!exec_lookupname(o, var, &name)) {
 		return 0;
 	}
@@ -192,15 +184,14 @@ exec_lookupfilt(struct exec_s *o, char *var, struct songfilt_s **res) {
 	return 1;
 }
 
-	/*
-	 * find the pointer to the songsx contained in 'var'
-	 * ('var' must be a reference)
-	 */
-
+/*
+ * find the pointer to the songsx contained in 'var' ('var' must be a
+ * reference)
+ */
 unsigned
-exec_lookupsx(struct exec_s *o, char *var, struct songsx_s **res) {
+exec_lookupsx(struct exec *o, char *var, struct songsx **res) {
 	char *name;	
-	struct songsx_s *t;
+	struct songsx *t;
 	if (!exec_lookupname(o, var, &name)) {
 		return 0;
 	}
@@ -213,25 +204,28 @@ exec_lookupsx(struct exec_s *o, char *var, struct songsx_s **res) {
 	return 1;
 }
 
-	/*
-	 * fill the event with the one referenced by 'var'
-	 * 'var' can be:
-	 * 	- { noff chan xxx yyy }
-	 * 	- { non chan xxx yyy }
-	 * 	- { ctl chan xxx yyy }
-	 * 	- { kat chan xxx yyy }
-	 * 	- { cat chan xxx }
-	 * 	- { pc chan xxx }
-	 * 	- { bend chan xxx }
-	 * where 'chan' is in the same as in lookupchan_getnum
-	 * and 'xxx' and 'yyy' are integers
-	 */
-
+/*
+ * fill the event with the one referenced by 'var'
+ * 'var' can be:
+ * 	- { noff chan xxx yyy }
+ * 	- { non chan xxx yyy }
+ * 	- { ctl chan xxx yyy }
+ * 	- { kat chan xxx yyy }
+ * 	- { cat chan xxx }
+ * 	- { pc chan xxx }
+ * 	- { bend chan xxx }
+ *	- { xctl chan xxx yyy }
+ *	- { xpc chan xxx yyy }
+ *	- { rpn chan xxx yyy }
+ *	- { nrpn chan xxx yyy }
+ * where 'chan' is in the same as in lookupchan_getnum
+ * and 'xxx' and 'yyy' are integers
+ */
 unsigned
-exec_lookupev(struct exec_s *o, char *name, struct ev_s *ev) {
-	struct var_s *arg;
-	struct data_s *d;
-	unsigned dev, ch;
+exec_lookupev(struct exec *o, char *name, struct ev *ev) {
+	struct var *arg;
+	struct data *d;
+	unsigned dev, ch, max, num;
 
 	arg = exec_varlookup(o, name);
 	if (!arg) {
@@ -259,47 +253,73 @@ exec_lookupev(struct exec_s *o, char *name, struct ev_s *ev) {
 	if (!data_list2chan(d, &dev, &ch)) {
 		return 0;
 	}
-	ev->data.voice.dev = dev;
-	ev->data.voice.ch = ch;
+	ev->dev = dev;
+	ev->ch = ch;
 	d = d->next;
-	if (!d || d->type != DATA_LONG || d->val.num < 0 || d->val.num > 127) {
-		cons_err("bad byte0 in event spec");
-		return 0;
-	}
-	if (ev->cmd == EV_BEND) {
-		EV_SETBEND(ev, d->val.num);
-	} else {
-		ev->data.voice.b0 = d->val.num;
-	}		
-	d = d->next;
-	if (ev->cmd != EV_PC && ev->cmd != EV_CAT && ev->cmd != EV_BEND) {
-		if (!d || d->type != DATA_LONG || d->val.num < 0 || d->val.num > 127) {
-			cons_err("bad byte1 in event spec");
+	if (ev->cmd == EV_XCTL || ev->cmd == EV_CTL) {
+		if (!d || !data_getctl(d, &num)) {
 			return 0;
 		}
-		ev->data.voice.b1 = d->val.num;
+		ev->ctl_num = num;
 	} else {
+		if (ev->cmd == EV_BEND || ev->cmd == EV_NRPN || ev->cmd == EV_RPN) {
+			max = EV_MAXFINE; 
+		} else {
+			max = EV_MAXCOARSE;
+		}
+		if (!d || d->type != DATA_LONG || d->val.num < 0 || d->val.num > max) {
+			cons_err("bad byte0 in event spec");
+			return 0;
+		}
+		ev->v0 = d->val.num;
+	}
+	d = d->next;
+	if (ev->cmd == EV_PC || ev->cmd == EV_CAT || ev->cmd == EV_BEND) {
 		if (d) {
 			cons_err("extra data in event spec");
 			return 0;
 		}
+	} else {
+		if (ev->cmd == EV_XPC || ev->cmd == EV_XCTL ||
+		    ev->cmd == EV_NRPN || ev->cmd == EV_RPN) {
+			max = EV_MAXFINE;
+		} else {
+			max = EV_MAXCOARSE;
+		}
+		if (!d || d->type != DATA_LONG || 
+		    d->val.num < 0 || d->val.num > max) {
+			cons_err("bad byte1 in event spec");
+			return 0;
+		}
+		ev->v1 = d->val.num;
+	}
+
+	/*
+	 * convert all CTL -> XCTL and PC -> XPC
+	 */
+	if (ev->cmd == EV_CTL) {
+		ev->cmd = EV_XCTL;
+		ev->ctl_val <<= 7;
+	}
+	if (ev->cmd == EV_PC) {
+		ev->cmd = EV_XPC;
+		ev->pc_bank = EV_UNDEF;
 	}
 	return 1;
 }
 
-	/* 
-	 * fill the evspec with the one referenced by var.
-	 * var is of this form
-	 * 	- { [type [chanrange [xxxrange [yyyrange]]]] }	
-	 * (brackets mean optionnal)
-	 */
-
+/* 
+ * fill the evspec with the one referenced by var.
+ * var is of this form
+ * 	- { [type [chanrange [xxxrange [yyyrange]]]] }	
+ * (brackets mean argument is optionnal)
+ */
 unsigned
-exec_lookupevspec(struct exec_s *o, char *name, struct evspec_s *e) {
-	struct var_s *arg;
-	struct data_s *d;
-	struct songchan_s *i;
-	unsigned lo, hi;
+exec_lookupevspec(struct exec *o, char *name, struct evspec *e) {
+	struct var *arg;
+	struct data *d;
+	struct songchan *i;
+	unsigned lo, hi, max;
 
 	arg = exec_varlookup(o, name);
 	if (!arg) {
@@ -311,28 +331,27 @@ exec_lookupevspec(struct exec_s *o, char *name, struct evspec_s *e) {
 		cons_err("list expected in event range spec");
 		return 0;
 	}
-
-	/* default match any event */
 	evspec_reset(e);
 
+	/*
+	 * find the event type
+	 */
 	d = d->val.list;
 	if (!d) {
-		/* empty list match anything */
-		return 1;
+		goto done;
 	}
 	if (d->type != DATA_REF ||
 	    !evspec_str2cmd(e, d->val.ref)) {
 		cons_err("bad status in event spec");
 		return 0;
-	}
-	
+	}	
+
+	/*
+	 * find the {device channel} pair
+	 */
 	d = d->next;
 	if (!d) {
-		/* 
-		 * match events of the given type (or 'any')
-		 * on any channel, with any value(s)
-		 */
-		return 1;
+		goto done;
 	}
 	if (d->type == DATA_REF) {
 		i = song_chanlookup(user_song, d->val.ref);
@@ -364,64 +383,118 @@ exec_lookupevspec(struct exec_s *o, char *name, struct evspec_s *e) {
 		}
 	}
 
+	/*
+	 * find the first parameter range
+	 */
 	d = d->next;
 	if (!d) {
-		/* match events of the given type (or 'any' 
-		 * with the given channel range but with any value(s) */ 
-		return 1;
+		goto done;
 	}
 	if (e->cmd == EVSPEC_ANY) {
 		goto toomany;
 	}
-	if (e->cmd == EVSPEC_BEND) {
-		if (!data_list2range(d, 0, EV_MAXBEND, &lo, &hi)) {
+	if ((e->cmd == EVSPEC_CTL || e->cmd == EV_XCTL) && d->type == DATA_REF) {
+		if (!data_getctl(d, &hi)) {
 			return 0;
 		}
-		e->b0_min = lo;
+		e->b0_min = hi;
 		e->b0_max = hi;
 	} else {
-		if (!data_list2range(d, 0, EV_MAXB0, &lo, &hi)) {
+		if (e->cmd == EVSPEC_BEND || e->cmd == EVSPEC_XCTL || 
+		    e->cmd == EVSPEC_NRPN || e->cmd == EVSPEC_RPN) {
+			max = EV_MAXFINE;
+		} else {
+			max = EV_MAXCOARSE;
+		}	
+		if (!data_list2range(d, 0, max, &lo, &hi)) {
 			return 0;
 		}
 		e->b0_min = lo;
 		e->b0_max = hi;
 	}
-	d = d->next;
-	if (!d) {
-		return 1;
-	}
-	if (e->cmd != EVSPEC_PC && 
-	    e->cmd != EVSPEC_CAT && 
-	    e->cmd != EVSPEC_BEND) {
-		if (!data_list2range(d, 0, EV_MAXB1, &lo, &hi)) {
-			return 0;
-		}
-		e->b1_min = lo;
-		e->b1_max = hi;
-		d = d->next;
-		if (!d) {
-			return 1;
-		}
-	}
-toomany:
-	cons_err("too many ranges in event spec");
-	return 0;				
-}
 
 	/*
-	 * print a data_s to the user console
+	 * find the second parameter range
 	 */
+	d = d->next;
+	if (!d) {
+		goto done;
+	}
+	if ((e->cmd == EVSPEC_PC || e->cmd == EVSPEC_CAT || 
+	     e->cmd == EVSPEC_BEND)) {
+		goto toomany;
+	}
+	if (e->cmd == EVSPEC_CTL || e->cmd == EVSPEC_NOTE) {
+		max = EV_MAXCOARSE;
+	} else {
+		max = EV_MAXFINE;
+	}
+	if (!data_list2range(d, 0, max, &lo, &hi)) {
+		return 0;
+	}
+	e->b1_min = lo;
+	e->b1_max = hi;
+	d = d->next;
+	if (d) {
+		goto toomany;
+	}
 
+ done:
+	/*
+	 * convert PC->XPC and CTL->XCTL
+	 */
+	if (e->cmd == EVSPEC_PC) {
+		e->cmd = EVSPEC_XPC;
+	} else if (e->cmd == EVSPEC_CTL) {
+		e->cmd = EVSPEC_XCTL;
+		e->b1_min =  (e->b1_min << 7) & 0x3fff;
+		e->b1_max = ((e->b1_max << 7) & 0x3fff) | 0x7f;
+	}
+	dbg_puts("evspec: ");
+	evspec_dbg(e);
+	dbg_puts("\n");
+	return 1;
+ toomany:
+	cons_err("too many ranges/values in event spec");
+	return 0;
+}
+
+/*
+ * find the (hi lo) couple for the 14bit controller
+ * number:
+ *	- an integer for 7bit controller
+ *	- a list of two integers (like '{hi lo}')
+ */
+unsigned
+exec_lookupctl(struct exec *o, char *var, unsigned *num) {
+	struct var *arg;
+	
+	arg = exec_varlookup(o, var);
+	if (!arg) {
+		dbg_puts("exec_lookupctl: no such var\n");
+		dbg_panic();
+	}
+	return data_getctl(arg->data, num);
+}
+
+/*
+ * print a struct data to the console
+ */
 void
-data_print(struct data_s *d) {
-	struct data_s *i;
+data_print(struct data *d) {
+	struct data *i;
 	
 	switch(d->type) {
 	case DATA_NIL:
 		textout_putstr(tout, "nil");
 		break;
 	case DATA_LONG:
-		textout_putlong(tout, d->val.num);
+		if (d->val.num < 0) {
+			textout_putstr(tout, "-");
+			textout_putlong(tout, -d->val.num);
+		} else {
+			textout_putlong(tout, d->val.num);
+		}
 		break;
 	case DATA_STRING:
 		textout_putstr(tout, "\"");
@@ -447,16 +520,14 @@ data_print(struct data_s *d) {
 	}
 }
 
-	/*
-	 * convert 2 integer lists to channels
-	 */
-	 
+/*
+ * convert 2 integer lists to channels
+ */
 unsigned
-data_num2chan(struct data_s *o, unsigned *res_dev, unsigned *res_ch) {
+data_num2chan(struct data *o, unsigned *res_dev, unsigned *res_ch) {
 	long dev, ch;
 
-	if (o->type == DATA_LIST ||
-	    o == NULL || 
+	if (o == NULL || 
 	    o->next == NULL || 
 	    o->next->next != NULL ||
 	    o->type != DATA_LONG || 
@@ -476,15 +547,44 @@ data_num2chan(struct data_s *o, unsigned *res_dev, unsigned *res_ch) {
 	return 1;
 }
 
-	/*
-	 * convert lists to channels, 'data' can be
-	 * 	- a reference to an existing songchan
-	 *	- a pair of integers '{ dev midichan }'
-	 */
-	 
+/*
+ * lookup the value of the given controller
+ */
 unsigned
-data_list2chan(struct data_s *o, unsigned *res_dev, unsigned *res_ch) {
-	struct songchan_s *i;
+exec_lookupval(struct exec *o, char *n, unsigned isfine, unsigned *r) {
+	struct var *arg;
+	unsigned max;
+
+	arg = exec_varlookup(o, n);
+	if (!arg) {
+		dbg_puts("exec_lookupval: no such var\n");
+		dbg_panic();
+	}
+	if (arg->data->type == DATA_NIL) {
+		*r = EV_UNDEF;
+		return 1;
+	} else if (arg->data->type == DATA_LONG) {
+		   max = isfine ? EV_MAXFINE : EV_MAXCOARSE;
+		   if (arg->data->val.num < 0 || arg->data->val.num > max) {
+			   cons_err("controller value out of range");
+			   return 0;
+		   }
+		   *r = arg->data->val.num;
+		   return 1;
+	} else {
+		cons_err("bad type of controller value\n");
+		return 0;
+	}
+}
+
+/*
+ * convert lists to channels, 'data' can be
+ * 	- a reference to an existing songchan
+ *	- a pair of integers '{ dev midichan }'
+ */
+unsigned
+data_list2chan(struct data *o, unsigned *res_dev, unsigned *res_ch) {
+	struct songchan *i;
 
 	if (o->type == DATA_LIST) {
 		return data_num2chan(o->val.list, res_dev, res_ch);
@@ -503,17 +603,14 @@ data_list2chan(struct data_s *o, unsigned *res_dev, unsigned *res_ch) {
 	}
 }
 
-
-	/*
-	 * convert a data_s to a pair of integers
-	 * data can be:
-	 * 	- a liste of 2 integers
-	 *	- a single integer (then min = max)
-	 */
-	 
-	 
+/*
+ * convert a struct data to a pair of integers
+ * data can be:
+ * 	- a liste of 2 integers
+ *	- a single integer (then min = max)
+ */
 unsigned
-data_list2range(struct data_s *d, unsigned min, unsigned max, 
+data_list2range(struct data *d, unsigned min, unsigned max, 
     unsigned *lo, unsigned *hi) {
     	if (d->type == DATA_LONG) {
 		*lo = *hi = d->val.num;
@@ -526,7 +623,7 @@ data_list2range(struct data_s *d, unsigned min, unsigned max,
 		} 
 		if (!d->next || d->next->next || 
 		    d->type != DATA_LONG || d->next->type != DATA_LONG) {
-			cons_err("exactly 0 ore 2 numbers expected in range spec");
+			cons_err("exactly 0 or 2 numbers expected in range spec");
 			return 0;
 		}
 		*lo = d->val.num;
@@ -542,15 +639,43 @@ data_list2range(struct data_s *d, unsigned min, unsigned max,
 	return 1;
 }
 
-	/*
-	 * check if the pattern in data (list of integers)
-	 * match the beggining of the given sysex
-	 */
-
+/*
+ * convert a list to bitmap of continuous controllers
+ */
 unsigned
-data_matchsysex(struct data_s *d, struct sysex_s *sx, unsigned *res) {
+data_list2ctlset(struct data *d, unsigned *res) {
+	unsigned ctlset;
+	
+	ctlset = 0;
+	while (d) {
+		if (d->type != DATA_LONG) {
+			cons_err("not-a-number in controller set");
+			return 0;
+		}
+		if (d->val.num < 0 || d->val.num >= 32) {
+			cons_err("controller number out of range 0..31");
+			return 0;
+		}
+		if (evctl_isreserved(d->val.num)) {
+			cons_erru(d->val.num, "controller number reserved");
+			return 0;
+		}
+		ctlset |= (1 << d->val.num);
+		d = d->next;	
+	}
+	*res = ctlset;
+	return 1;
+}
+
+
+/*
+ * check if the pattern in data (list of integers)
+ * match the beggining of the given sysex
+ */
+unsigned
+data_matchsysex(struct data *d, struct sysex *sx, unsigned *res) {
 	unsigned i;
-	struct chunk_s *ck;
+	struct chunk *ck;
 	
 	i = 0;
 	ck = sx->first;
@@ -580,33 +705,44 @@ data_matchsysex(struct data_s *d, struct sysex_s *sx, unsigned *res) {
 	return 1;
 }
 
-
-/* ---------------------------------------- interpreter functions --- */
-
-	/* XXX: for testing */
-
+/*
+ * convert a 'struct data' to a controller number
+ */
 unsigned
-user_func_ev(struct exec_s *o, struct data_s **r) {
-	struct evspec_s ev;
-	if (!exec_lookupevspec(o, "ev", &ev)) {
+data_getctl(struct data *d, unsigned *num) {
+    	if (d->type == DATA_LONG) {
+		if (d->val.num < 0 || d->val.num > EV_MAXCOARSE) {
+			cons_err("7bit ctl number out of bounds");
+			return 0;
+		}
+		if (evctl_isreserved(d->val.num)) {
+			cons_err("controller is reserved for bank, rpn, nrpn");
+			return 0;
+		}
+		*num = d->val.num;
+	} else if (d->type == DATA_REF) {
+		if (!evctl_lookup(d->val.ref, num)) {
+			cons_errs(d->val.ref, "no such controller\n");
+			return 0;
+		}
+	} else {
+		cons_err("number or identifier expected in ctl spec");
 		return 0;
 	}
-	evspec_dbg(&ev);
-	dbg_puts("\n");
 	return 1;
 }
 
+/* ---------------------------------------- interpreter functions --- */
 
 unsigned
-user_func_panic(struct exec_s *o, struct data_s **r) {
+user_func_panic(struct exec *o, struct data **r) {
 	dbg_panic();
 	/* not reached */
 	return 0;
 }
 
-
 unsigned
-user_func_debug(struct exec_s *o, struct data_s **r) {
+user_func_debug(struct exec *o, struct data **r) {
 	char *flag;
 	long value;
 	
@@ -620,6 +756,8 @@ user_func_debug(struct exec_s *o, struct data_s **r) {
 		filt_debug = value;
 	} else if (str_eq(flag, "song")) {
 		song_debug = value;
+	} else if (str_eq(flag, "mem")) {
+		mem_debug = value;
 	} else {
 		cons_err("debug: unknuwn debug-flag");
 		return 0;
@@ -628,7 +766,7 @@ user_func_debug(struct exec_s *o, struct data_s **r) {
 }
 
 unsigned
-user_func_exec(struct exec_s *o, struct data_s **r) {
+user_func_exec(struct exec *o, struct data **r) {
 	char *filename;		
 	if (!exec_lookupstring(o, "filename", &filename)) {
 		return 0;
@@ -636,10 +774,9 @@ user_func_exec(struct exec_s *o, struct data_s **r) {
 	return exec_runfile(o, filename);
 }
 
-
 unsigned
-user_func_print(struct exec_s *o, struct data_s **r) {
-	struct var_s *arg;
+user_func_print(struct exec *o, struct data **r) {
+	struct var *arg;
 	arg = exec_varlookup(o, "value");
 	if (!arg) {
 		dbg_puts("user_func_print: 'value': no such param\n");
@@ -652,25 +789,25 @@ user_func_print(struct exec_s *o, struct data_s **r) {
 }
 
 unsigned
-user_func_info(struct exec_s *o, struct data_s **r) {
+user_func_info(struct exec *o, struct data **r) {
 	exec_dumpprocs(o);
 	exec_dumpvars(o);
 	return 1;
 }
 
 unsigned
-user_func_metroswitch(struct exec_s *o, struct data_s **r) {
+user_func_metroswitch(struct exec *o, struct data **r) {
 	long onoff;
 	if (!exec_lookuplong(o, "onoff", &onoff)) {
 		return 0;
 	}	
-	user_song->metro_enabled = onoff;
+	user_song->metro.enabled = onoff;
 	return 1;
 }
 
 unsigned
-user_func_metroconf(struct exec_s *o, struct data_s **r) {
-	struct ev_s evhi, evlo;
+user_func_metroconf(struct exec *o, struct data **r) {
+	struct ev evhi, evlo;
 	if (!exec_lookupev(o, "eventhi", &evhi) ||
 	    !exec_lookupev(o, "eventlo", &evlo)) {
 		return 0;
@@ -679,50 +816,50 @@ user_func_metroconf(struct exec_s *o, struct data_s **r) {
 		cons_err("note-on event expected");
 		return 0;
 	}
-	user_song->metro_hi = evhi;
-	user_song->metro_lo = evlo;
+	user_song->metro.hi = evhi;
+	user_song->metro.lo = evlo;
 	return 1;
 }
 
 unsigned
-user_func_shut(struct exec_s *o, struct data_s **r) {
+user_func_shut(struct exec *o, struct data **r) {
 	unsigned i;
-	struct ev_s ev;
-	struct mididev_s *dev;
+	struct ev ev;
+	struct mididev *dev;
 
-	mux_init(NULL, NULL);
+	mux_open(NULL, NULL);
 
 	for (dev = mididev_list; dev != NULL; dev = dev->next) {
 		for (i = 0; i < EV_MAXCH; i++) {
-			ev.cmd = EV_CTL;
-			ev.data.voice.dev = dev->unit;		
-			ev.data.voice.ch = i;
-			ev.data.voice.b0 = 121;
-			ev.data.voice.b1 = 0;
+			ev.cmd = EV_XCTL;
+			ev.dev = dev->unit;		
+			ev.ch = i;
+			ev.ctl_num = 121;
+			ev.ctl_val = 0;
 			mux_putev(&ev);
-			ev.cmd = EV_CTL;		
-			ev.data.voice.dev = dev->unit;		
-			ev.data.voice.ch = i;
-			ev.data.voice.b0 = 123;
-			ev.data.voice.b1 = 0;
+			ev.cmd = EV_XCTL;		
+			ev.dev = dev->unit;		
+			ev.ch = i;
+			ev.ctl_num = 123;
+			ev.ctl_val = 0;
 			mux_putev(&ev);
 			ev.cmd = EV_BEND;
-			ev.data.voice.dev = dev->unit;		
-			ev.data.voice.ch = i;
-			ev.data.voice.b0 = EV_BEND_DEFAULTLO;
-			ev.data.voice.b1 = EV_BEND_DEFAULTHI;
+			ev.dev = dev->unit;		
+			ev.ch = i;
+			ev.bend_val = EV_BEND_DEFAULT;
+			mux_putev(&ev);
 		}
 	}
 	
 	mux_flush();
-	mux_done();
+	mux_close();
 	return 1;
 }
 
 unsigned
-user_func_sendraw(struct exec_s *o, struct data_s **r) {
-	struct var_s *arg;
-	struct data_s *i;
+user_func_sendraw(struct exec *o, struct data **r) {
+	struct var *arg;
+	struct data *i;
 	unsigned char byte;
 	long device;
 	
@@ -749,29 +886,34 @@ user_func_sendraw(struct exec_s *o, struct data_s **r) {
 		}
 	}
 	
-	mux_init(NULL, NULL);
+	mux_open(NULL, NULL);
 	for (i = arg->data->val.list; i != NULL; i = i->next) {
 		byte = i->val.num;
 		mux_sendraw(device, &byte, 1);
 	}
 	mux_flush();
-	mux_done();
+	mux_close();
 	return 1;
 }
 
-
-void
+unsigned
 user_mainloop(void) {
-	struct parse_s *parse;
-	struct exec_s *exec;
-	struct node_s *root;
-	struct data_s *data;
+	struct parse *parse;
+	struct exec *exec;
+	struct node *root;
+	struct data *data;
+	unsigned result, exitcode;
 	
+	/* 
+	 * create the project (ie the song) and
+	 * the execution environment of the interpreter
+	 */
 	user_song = song_new();
 	exec = exec_new();
 
-	exec_newbuiltin(exec, "ev", user_func_ev, 
-			name_newarg("ev", NULL));
+	/*
+	 * register built-in functions 
+	 */
 	exec_newbuiltin(exec, "print", user_func_print, 
 			name_newarg("value", NULL));
 	exec_newbuiltin(exec, "exec", user_func_exec, 
@@ -829,12 +971,22 @@ user_mainloop(void) {
 			name_newarg("from", 
 			name_newarg("amount", 
 			name_newarg("quantum", NULL)))));
+	exec_newbuiltin(exec, "trackmerge", user_func_trackmerge,
+			name_newarg("source", 
+			name_newarg("dest", NULL)));
 	exec_newbuiltin(exec, "trackquant", user_func_trackquant, 
 			name_newarg("trackname", 
 			name_newarg("from", 
 			name_newarg("amount", 
 			name_newarg("rate", 
 			name_newarg("quantum", NULL))))));
+	exec_newbuiltin(exec, "tracktransp", user_func_tracktransp, 
+			name_newarg("trackname", 
+			name_newarg("from", 
+			name_newarg("amount", 
+			name_newarg("halftones",
+			name_newarg("quantum", 
+			name_newarg("evspec", NULL)))))));
 	exec_newbuiltin(exec, "tracksetmute", user_func_tracksetmute,
 			name_newarg("trackname",
 			name_newarg("muteflag", NULL)));
@@ -870,6 +1022,9 @@ user_mainloop(void) {
 	exec_newbuiltin(exec, "chanconfev", user_func_chanconfev,
 			name_newarg("channame",
 			name_newarg("event", NULL)));
+	exec_newbuiltin(exec, "chanunconfev", user_func_chanunconfev,
+			name_newarg("channame",
+			name_newarg("evspec", NULL)));
 	exec_newbuiltin(exec, "chansetcurinput", user_func_chansetcurinput, 
 			name_newarg("channame", 
 			name_newarg("inputchan", NULL)));
@@ -1021,6 +1176,9 @@ user_mainloop(void) {
 	exec_newbuiltin(exec, "songgetunit", user_func_songgetunit, NULL);
 	exec_newbuiltin(exec, "songsetunit", user_func_songsetunit, 
 			name_newarg("tics_per_unit", NULL));
+	exec_newbuiltin(exec, "songgetfactor", user_func_songgetfactor, NULL);
+	exec_newbuiltin(exec, "songsetfactor", user_func_songsetfactor, 
+			name_newarg("tempo_factor", NULL));
 	exec_newbuiltin(exec, "songgetcurpos", user_func_songgetcurpos, NULL);
 	exec_newbuiltin(exec, "songsetcurpos", user_func_songsetcurpos, 
 			name_newarg("measure", NULL));
@@ -1071,6 +1229,17 @@ user_mainloop(void) {
 			name_newarg("from", 
 			name_newarg("amount", NULL)));
 	exec_newbuiltin(exec, "songtimeinfo", user_func_songtimeinfo, NULL);
+	exec_newbuiltin(exec, "ctlconf", user_func_ctlconf, 
+			name_newarg("name", 
+			name_newarg("ctl", 
+			name_newarg("defval", NULL))));
+	exec_newbuiltin(exec, "ctlconfx", user_func_ctlconf, 
+			name_newarg("name", 
+			name_newarg("ctl", 
+			name_newarg("defval", NULL))));
+	exec_newbuiltin(exec, "ctlunconf", user_func_ctlunconf, 
+			name_newarg("name", NULL));
+	exec_newbuiltin(exec, "ctlinfo", user_func_ctlinfo, NULL);
 
 	exec_newbuiltin(exec, "metroswitch", user_func_metroswitch, 
 			name_newarg("onoff", NULL));
@@ -1084,7 +1253,8 @@ user_mainloop(void) {
 
 	exec_newbuiltin(exec, "devattach", user_func_devattach,
 			name_newarg("unit", 
-			name_newarg("path", NULL)));
+			name_newarg("path", 
+			name_newarg("mode", NULL))));
 	exec_newbuiltin(exec, "devdetach", user_func_devdetach,
 			name_newarg("unit", NULL));
 	exec_newbuiltin(exec, "devlist", user_func_devlist, NULL);
@@ -1099,47 +1269,83 @@ user_mainloop(void) {
 			name_newarg("tics_per_unit", NULL)));
 	exec_newbuiltin(exec, "devinfo", user_func_devinfo,
 			name_newarg("unit", NULL));
+	exec_newbuiltin(exec, "devixctl", user_func_devixctl,
+			name_newarg("unit", 
+			name_newarg("ctlset", NULL)));
+	exec_newbuiltin(exec, "devoxctl", user_func_devoxctl,
+			name_newarg("unit", 
+			name_newarg("ctlset", NULL)));
 
+	/*
+	 * run the user startup script: $HOME/.midishrc or /etc/midishrc
+	 */
 	if (!user_flag_batch) {
 		exec_runrcfile(exec);
 	}
 
+	/*
+	 * create the parser and start parsing standard input
+	 */
 	parse = parse_new(NULL);
 	if (parse == NULL) {
-		return;
+		return 0;
 	}
 
 	root = NULL;
 	data = NULL;
 	for (;;) {
-		if (parse_getsym(parse)) {
-			/* at this stage no lexical error */
-			if (parse->lex.id == TOK_EOF) {
-				/* end-of-file (user quit) */
-				break;
-			}
-			parse_ungetsym(parse);
-			if (parse_line(parse, &root)) {
-				/* at this stage no parse error */
-				if (node_exec(root, exec, &data) == RESULT_OK) {
-					/* at this stage no exec error */
-					node_delete(root);
-					root = NULL;
-					continue;
-				}
-			}
+		/*
+		 * print mem_alloc() and mem_free() stats, useful to
+		 * track memory leaks
+		 */
+		mem_stats();
+		
+		/*
+		 * parse a block
+		 */
+		if (!parse_getsym(parse)) {
+			goto err;
+		}
+		if (parse->lex.id == TOK_EOF) {
+			/* end-of-file (user quit) */
+			exitcode = 1;
+			break;
+		}
+		parse_ungetsym(parse);
+		if (!parse_line(parse, &root)) {
 			node_delete(root);
 			root = NULL;
+			goto err;
 		}
-		/* error */
+		
+		/* 
+		 * at this stage no parse error, execute the tree
+		 */
+		result = node_exec(root, exec, &data);
+		node_delete(root);
+		root = NULL;
+		if (result == RESULT_OK) {
+			continue;
+		}
+		if (result == RESULT_EXIT) {
+			exitcode = 1;	/* 1 means success */
+			break;
+		}
+		
+	err:	
+		/* 
+		 * in batch mode stop on the first error
+		 */
 		if (user_flag_batch) {
+			exitcode = 0;	/* 0 means failure */
 			break;
 		}
 	}
-
+	
 	parse_delete(parse);
 	exec_delete(exec);
 	song_delete(user_song);
 	user_song = NULL;	
+	return exitcode;
 }
 
